"""
MIT License

Copyright (c) 2018 Ross Wilson
Copyright (c) 2024, Santiago Peñate Vera

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Well, in the OpenStreetMap world[1], a 'slippy map' is a browser map view
served by a tile server that can be panned and zoomed in the same way as
popularised by Google maps.  Such a map feels 'slippery', I guess.

Rather than 'slippy' I went for the slightly more formal 'pySlip' since the
thing is written in Python and therefore must have the obligatory 'py' prefix.

Even though this was originally written for a geographical application, the
*underlying* system only assumes a cartesian 2D coordinate system.  The tile
source must translate between the underlying coordinates and whatever coordinate
system the tiles use.  So pySlip could be used to present a game map, 2D CAD
view, etc, as well as Mercator tiles provided either locally from the filesystem
or from the internet (OpenStreetMap, for example).

[1] http://wiki.openstreetmap.org/index.php/Slippy_Map

Some semantics:
    map   the whole map
    view  is the view of the map through the widget
          (view may be smaller than map, or larger)

xgeo: longitude
ygeo: latitude
"""
from __future__ import annotations
from typing import List, Union, Tuple, Callable, TYPE_CHECKING
from enum import Enum
from PySide6.QtCore import Qt, QEvent, QPointF
from PySide6.QtGui import (QPainter, QColor, QPixmap, QCursor,
                           QMouseEvent, QKeyEvent, QWheelEvent,
                           QResizeEvent, QEnterEvent, QPaintEvent, QDragEnterEvent, QDragMoveEvent, QDropEvent)
from PySide6.QtWidgets import (QSizePolicy, QWidget, QGraphicsScene, QGraphicsView, QStackedLayout,
                               QGraphicsSceneMouseEvent, QGraphicsItem, QLabel, QGraphicsProxyWidget)

from VeraGrid.Gui.Diagrams.MapWidget.Tiles.tiles import Tiles

if TYPE_CHECKING:
    from VeraGrid.Gui.Diagrams.MapWidget.grid_map_widget import GridMapWidget


class Place(Enum):
    """
    places to draw in the map
    """

    Center = "cc"
    NorthWest = "nw"
    CenterNorth = "cn"
    NorthEast = "ne"
    CenterEast = "ce"
    SouthEast = "se"
    CenterSouth = "cs"
    SouthWest = "sw"
    CenterWest = "cw"


class MapDiagramScene(QGraphicsScene):
    """
    CustomScene
    """

    def __init__(self, parent: "MapWidget" = None) -> None:
        super().__init__(parent)

        self.setItemIndexMethod(QGraphicsScene.ItemIndexMethod.BspTreeIndex)  # For efficient item indexing

    def mousePressEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """

        :param event:
        :return:
        """
        # print(f"Scene pressed at {event.scenePos()}")
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """

        :param event:
        :return:
        """
        # print(f"Scene released at {event.scenePos()}")
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        """

        :param event:
        :return:
        """
        super().mouseMoveEvent(event)


class MapView(QGraphicsView):
    """
    MapView
    """

    def __init__(self,
                 scene: QGraphicsScene,
                 map_widget: "MapWidget"):
        """

        :param scene:
        :param map_widget:
        """
        super().__init__(scene)

        self._scene: QGraphicsScene = scene

        self.map_widget = map_widget
        self.setStyleSheet("QGraphicsView { border: none; }")

        # Create a QLabel
        self.attribution_label = QLabel("Bottom Left Label")
        self.attribution_label.setStyleSheet("background-color: rgba(0, 0, 0, 0);"
                                             "color: rgba(150, 150, 150, 180);"
                                             "font-size:9pt")  # Semi-transparent yellow

        # Create a QGraphicsProxyWidget for the QLabel
        self.label_proxy_widget = QGraphicsProxyWidget()
        self.label_proxy_widget.setWidget(self.attribution_label)
        self.label_proxy_widget.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIgnoresTransformations)
        self.update_label_position()

        # Add the proxy widget to the scene
        self._scene.addItem(self.label_proxy_widget)

        self.mouse_x = None
        self.mouse_y = None

        self.diagram_w = 25000
        self.diagram_H = 25000

        # self.in_item = False  # looks that it is written but never used
        self.pressed = False
        self.disable_move = False

        # updated later
        self.view_width = self.width()
        self.view_height = self.height()

        # Set initial zoom level (change the values as needed)
        initial_zoom_factor = 1.0
        self.schema_zoom = 1.0

        self.drag_mode = QGraphicsView.DragMode.ScrollHandDrag
        self.setDragMode(self.drag_mode)

        self.scale(initial_zoom_factor, initial_zoom_factor)

        self.setRubberBandSelectionMode(Qt.ItemSelectionMode.IntersectsItemShape)

        self.setRenderHints(QPainter.RenderHint.Antialiasing | QPainter.RenderHint.SmoothPixmapTransform)

    def set_notice(self, val: str):
        """

        :param val:
        :return:
        """
        self.attribution_label.setText(val)

    def selected_items(self) -> List[QGraphicsItem]:
        """
        Get the selected items
        :return:
        """
        return self._scene.selectedItems()

    def mousePressEvent(self, event: QMouseEvent):
        """

        :param event:
        :return:
        """
        self.map_widget.mousePressEvent(event)
        self.pressed = True
        self.disable_move = False

        # By pressing ctrl while dragging, we can move the grid
        if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            self.drag_mode = QGraphicsView.DragMode.RubberBandDrag
            self.map_widget.block_movement = True
        else:
            self.drag_mode = QGraphicsView.DragMode.ScrollHandDrag
            self.map_widget.block_movement = False

        self.setDragMode(self.drag_mode)

        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        """

        :param event:
        :return:
        """
        self.map_widget.mouseReleaseEvent(event)
        self.pressed = False
        self.disable_move = True
        super().mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        """

        :param event:
        :return:
        """
        self.map_widget.mouseDoubleClickEvent(event)
        super().mouseDoubleClickEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent):
        """

        :param event:
        :return:
        """
        if not self.disable_move:
            self.map_widget.mouseMoveEvent(event)
            self.center_schema()

        super().mouseMoveEvent(event)
        self.update_label_position()

    def keyPressEvent(self, event: QKeyEvent):
        """

        :param event:
        :return:
        """
        self.map_widget.keyPressEvent(event)

    def keyReleaseEvent(self, event: QKeyEvent):
        """

        :param event:
        :return:
        """
        self.map_widget.keyReleaseEvent(event)

    def wheelEvent(self, event: QWheelEvent):
        """

        :param event:
        :return:
        """
        zoom_0 = self.map_widget.level

        self.mouse_x = event.position().x()
        self.mouse_y = event.position().y()

        if event.angleDelta().y() > 0:
            new_level = zoom_0 + 1
        else:
            new_level = zoom_0 - 1

        if self.map_widget.tile_src.level_in_range(new_level):

            val = self.map_widget.set_zoom_level(level=new_level,
                                                 view_x=self.mouse_x,
                                                 view_y=self.mouse_y)

            if val:
                if event.angleDelta().y() > 0:
                    self.schema_zoom = self.schema_zoom * self.map_widget.zoom_factor
                    self.scale(
                        self.map_widget.zoom_factor,
                        self.map_widget.zoom_factor
                    )
                else:
                    self.schema_zoom = self.schema_zoom / self.map_widget.zoom_factor
                    self.scale(
                        1.0 / self.map_widget.zoom_factor,
                        1.0 / self.map_widget.zoom_factor
                    )
            else:
                # revert to the previous zoom
                self.map_widget.set_zoom_level(level=zoom_0,
                                               view_x=self.mouse_x,
                                               view_y=self.mouse_y)

            self.map_widget.wheelEvent(event)
            self.center_schema()
            self.update_label_position()

        else:
            print(f"Zoom {new_level} out of range...")

    def dragEnterEvent(self, event: QDragEnterEvent):
        """

        :param event:
        :return:
        """
        if event.mimeData().hasFormat('component/name'):
            event.accept()

    def dragMoveEvent(self, event: QDragMoveEvent):
        """

        :param event:
        :return:
        """
        pass

    def dropEvent(self, event: QDropEvent):
        """

        :param event:
        :return:
        """
        super().dropEvent(event)
        self.map_widget.dropEvent(event)

    def resizeEvent(self, event: QResizeEvent = None):
        """
        Widget resized, recompute some state.
        """

        super().resizeEvent(event)
        self.map_widget.resizeEvent(event=event)

    def update_label_position(self):
        """
        Updates the position of the label to the bottom-left corner of the viewport.
        """
        view_width = self.viewport().width()
        view_height = self.viewport().height()

        # Set position relative to the bottom-left corner of the viewport
        self.label_proxy_widget.setPos(self.mapToScene(0, view_height - self.attribution_label.height()))

    def set_size_diagram(self) -> None:
        """

        :return:
        """
        # new widget size
        self.view_width = self.width()
        self.view_height = self.height()

    def set_scene_rect_diagram(self) -> None:
        """

        :return:
        """
        used_width = self.diagram_w
        used_height = self.diagram_H

        xToDiagram = -(used_width / 2)
        yToDiagram = -(used_height / 2)

        # Adjust the scene rect if needed
        self.setSceneRect(xToDiagram, yToDiagram, used_width, used_height)

        self.center_schema()

    def to_lat_lon(self, x: float, y: float) -> Tuple[float, float]:
        """
        Convert x, y position in the map to latitude and longitude
        :param x: x position in pixels
        :param y: y position in pixels
        :return: latitude, longitude
        """

        ix, iy = self.map_widget.geo_to_view(longitude=0, latitude=0)

        x2 = (x * self.schema_zoom) + ix
        y2 = (y * self.schema_zoom) + iy

        lon, lat = self.map_widget.view_to_geo_float(xview=x2, yview=y2)

        return lat, lon

    def to_x_y(self, lat: float, lon: float) -> Tuple[float, float]:
        """

        :param lat: latitude (deg)
        :param lon: longitude (deg)
        :return: x, y in the map
        """

        ix, iy = self.map_widget.geo_to_view(longitude=0.0, latitude=0.0)

        x, y = self.map_widget.geo_to_view(longitude=lon, latitude=lat)

        x = (x - ix) / self.schema_zoom
        y = (y - iy) / self.schema_zoom

        return x, y

    def center_schema(self) -> None:
        """
        This function centers the schema relative to the map according to lat. and long.
        """

        he = self.map_widget.height() / 2.0
        wi = self.map_widget.width() / 2.0

        lon, lat = self.map_widget.view_to_geo_float(xview=wi, yview=he)

        sx, sy = self.to_x_y(lat=lat, lon=lon)

        point = QPointF(sx - 5 / self.schema_zoom, sy - 5 / self.schema_zoom)

        self.map_widget.view.centerOn(point)

    def get_selected(self) -> List[QGraphicsItem]:
        """
        Get the selection
        :return:
        """
        return self._scene.selectedItems()


class MapWidget(QWidget):
    """
    Map widget
    """

    def __init__(self,
                 parent: Union[None, QWidget],
                 tile_src: Tiles,
                 start_level: int,
                 editor: GridMapWidget,
                 zoom_callback: Callable[[int], None],
                 position_callback: Callable[[float, float, int, int], None]):
        """
        Initialize the widget.
        :param parent: the GUI parent widget
        :param tile_src: a Tiles object, source of tiles
        :param start_level: level to initially display
        :param zoom_callback: zoom change callback function
        :param position_callback: position change callback function
        """

        super().__init__(parent)  # inherit all parent object setup

        # this is where you draw
        self.diagram_scene = MapDiagramScene(self)

        # pointer to the editor
        self.editor: GridMapWidget = editor

        # Create a layout for the view
        self.layout = QStackedLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(0)
        self.setStyleSheet("background-color: transparent;")

        # the view is the transparent layer used to draw stuff
        self.view = MapView(scene=self.diagram_scene, map_widget=self)
        self.view.setBackgroundBrush(Qt.GlobalColor.transparent)

        # -------------------------------------------------------------------------
        # Internal vars
        # -------------------------------------------------------------------------
        # remember the tile source object
        self._tile_src: Tiles = tile_src.copy()
        self._tile_src.setCallback(self.on_tile_available)
        self.view.set_notice(val=self._tile_src.attribution_string)

        # the tile coordinates
        self.level: int = start_level

        # callbacks
        self.zoom_callback: Callable[[int], None] = zoom_callback
        self.position_callback: Callable[[float, float, int, int], None] = position_callback

        # define position and tile coords of the "key" tile
        self.key_tile_left = 0  # tile coordinates of key tile
        self.key_tile_top = 0
        self.key_tile_x_offset = 0  # view coordinates of key tile wrt view
        self.key_tile_y_offset = 0

        # we keep track of the cursor coordinates if cursor on map
        self.mouse_x: int = 0
        self.mouse_y: int = 0

        # state variables holding mouse buttons state
        self.left_button_down: bool = False
        self.mid_button_down: bool = False
        self.right_button_down: bool = False
        self._block_movement: bool = False

        # keyboard state variables
        self.shift_down = False
        self.zoom_factor = 2

        # when dragging, remember the initial start point
        self.start_drag_x: int = 0
        self.start_drag_y: int = 0

        self.view_llon = 0
        self.view_rlon = 0
        self.view_blat = 0
        self.view_tlat = 0

        # some cursors
        self.standard_cursor = QCursor(Qt.CursorShape.ArrowCursor)
        self.box_select_cursor = QCursor(Qt.CursorShape.CrossCursor)
        self.wait_cursor = QCursor(Qt.CursorShape.WaitCursor)
        self.drag_cursor = QCursor(Qt.CursorShape.OpenHandCursor)

        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.setMinimumSize(self.tile_width, self.tile_height)

        self.setMouseTracking(True)
        self.setEnabled(True)  # to receive key events?

        self.default_cursor = self.standard_cursor
        self.setCursor(self.standard_cursor)

        # do a "resize" after this function
        self.go_to_level_and_position(level=6, longitude=0, latitude=40)
        self.go_to_level_and_position(level=7, longitude=0, latitude=40)
        self.go_to_level_and_position(level=6, longitude=0, latitude=40)

        # add the widgets in a layered manner
        # self.layout.addWidget(self.notice_widget)
        self.layout.addWidget(self.view)  # Add the QGraphicsView to the layout

        self.setLayout(self.layout)  # Set the layout for the MapWidget

    @property
    def block_movement(self):
        return self._block_movement

    @block_movement.setter
    def block_movement(self, value: bool):
        self._block_movement = bool(value)

    @property
    def tile_src(self) -> Tiles:
        """
        Get the current tile source
        :return: Tiles
        """
        return self._tile_src

    @tile_src.setter
    def tile_src(self, tile_src: Tiles):
        """
        Set the current tile source
        :param tile_src: Tiles
        """

        level, longitude, latitude = self.get_level_and_position()

        if tile_src.tile_set_name != self._tile_src.tile_set_name:  # avoid changing tile sets to themselves
            self._tile_src: Tiles = tile_src.copy()
            self._tile_src.setCallback(self.on_tile_available)
            self.view.set_notice(val=self._tile_src.attribution_string)

            if self.GotoLevel(level):
                self.go_to_level_and_position(level=level, longitude=longitude, latitude=latitude)
                self.view.center_schema()
            else:
                while abs(self.view.schema_zoom - 0.015625) > 0.00001:
                    self.view.schema_zoom = self.view.schema_zoom / self.view.map_widget.zoom_factor
                    self.view.scale(1.0 / self.view.map_widget.zoom_factor, 1.0 / self.view.map_widget.zoom_factor)
                self.go_to_level_and_position(level=0, longitude=0, latitude=0)

    @property
    def max_level(self):
        """

        :return:
        """
        return self.tile_src.max_level

    @property
    def min_level(self):
        """

        :return:
        """
        return self.tile_src.min_level

    @property
    def tile_width(self):
        """

        :return:
        """
        return self.tile_src.tile_size_x

    @property
    def tile_height(self):
        """

        :return:
        """
        return self.tile_src.tile_size_y

    @property
    def num_tiles_x(self):
        """

        :return:
        """
        return self.tile_src.num_tiles_x

    @property
    def num_tiles_y(self):
        """

        :return:
        """
        return self.tile_src.num_tiles_y

    @property
    def map_width(self):
        """

        :return:
        """
        return self.num_tiles_x * self.tile_width  # virtual map width

    @property
    def map_height(self):
        """

        :return:
        """
        return self.num_tiles_y * self.tile_height  # virtual map height

    @property
    def view_width(self):
        """

        :return:
        """
        return self.width()

    @property
    def view_height(self):
        """

        :return:
        """
        return self.height()

    def on_tile_available(self, level: int, x: float, y: float, image: QPixmap, error: bool):
        """
        Called when a new 'net tile is available.

        level  the level the tile is for
        x, y   tile coordinates of the tile
        image  the tile image data
        error  True if there was an error

        We have enough information to redraw a specific tile,
        but we just redraw the widget.
        """

        self.update()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """
        Mouse button pressed.
        :param event:
        :return:
        """
        super().mousePressEvent(event)

        b = event.button()
        if b == Qt.MouseButton.NoButton:
            pass

        elif b == Qt.MouseButton.LeftButton:
            self.left_button_down = True
            self.editor.object_editor_table.setModel(None)

        elif b == Qt.MouseButton.MiddleButton:
            self.mid_button_down = True

        elif b == Qt.MouseButton.RightButton:
            self.right_button_down = True

        else:
            pass

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        """
        Mouse button was released.

        event.x & event.y  view coords when released

        Could be end of a drag or point or box selection.  If it's the end of
        a drag we don't do a lot.  If a selection we process that.
        """
        pos = event.position()
        x = pos.x()
        y = pos.y()

        # cursor back to normal in case it was a box select
        self.setCursor(self.default_cursor)

        b = event.button()
        if b == Qt.MouseButton.NoButton:
            pass
        elif b == Qt.MouseButton.LeftButton:
            self.left_button_down = False

            if self.start_drag_x is None:
                # not dragging, possible point selection
                # get click point in view & global coords
                self.view_to_geo(x, y)

            # turn off dragging, if we were
            self.start_drag_x = self.start_drag_y = None

            longitude, latitude = self.view_to_geo(x, y)
            self.position_callback(latitude, longitude, x, y)

        elif b == Qt.MouseButton.MiddleButton:
            self.mid_button_down = False

        elif b == Qt.MouseButton.RightButton:
            self.right_button_down = False

        else:
            pass
            # print('mouseReleaseEvent: unknown button')

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        """

        :param event:
        :return:
        """
        b = event.button()
        if b == Qt.MouseButton.NoButton:
            pass
        elif b == Qt.MouseButton.LeftButton:
            pass
        elif b == Qt.MouseButton.MiddleButton:
            pass
        elif b == Qt.MouseButton.RightButton:
            pass
        else:
            pass
            # print('mouseDoubleClickEvent: unknown button')

    def mouseMoveEvent(self, event: QMouseEvent):
        """
        Handle a mouse move event.

        If left mouse down, either drag the map or start a box selection.
        If we are off the map, ensure self.mouse_x, etc, are None.
        """
        if not self.block_movement:
            pos = event.position()
            x = pos.x()
            y = pos.y()

            mouse_geo = self.view_to_geo(x, y)

            # update remembered mouse position in case of zoom
            self.mouse_x = self.mouse_y = None
            if mouse_geo:
                self.mouse_x = x
                self.mouse_y = y

            if self.left_button_down:

                # we are dragging
                if self.start_drag_x is None:
                    # start of drag, set drag state
                    self.start_drag_x = x
                    self.start_drag_y = y

                # we don't move much - less than a tile width/height
                # drag the key tile in the X direction
                delta_x = self.start_drag_x - x
                self.key_tile_x_offset -= delta_x

                if self.key_tile_x_offset < -self.tile_width:  # too far left
                    self.key_tile_x_offset += self.tile_width
                    self.key_tile_left += 1

                if self.key_tile_x_offset > 0:  # too far right
                    self.key_tile_x_offset -= self.tile_width
                    self.key_tile_left -= 1

                # drag the key tile in the Y direction
                delta_y = self.start_drag_y - y
                self.key_tile_y_offset -= delta_y

                if self.key_tile_y_offset < -self.tile_height:  # too far up
                    self.key_tile_y_offset += self.tile_height
                    self.key_tile_top += 1

                if self.key_tile_y_offset > 0:  # too far down
                    self.key_tile_y_offset -= self.tile_height
                    self.key_tile_top -= 1

                # set key tile stuff so update() shows drag
                self.rectify_key_tile()

                # get ready for more drag
                self.start_drag_x = x
                self.start_drag_y = y

                self.update()  # force a repaint

    def keyPressEvent(self, event: QKeyEvent):
        """Capture a key press."""

        if event.key() == Qt.Key.Key_Shift:
            self.shift_down = True
            self.default_cursor = self.box_select_cursor
            self.setCursor(self.default_cursor)

        event.accept()

    def keyReleaseEvent(self, event: QKeyEvent):
        """Capture a key release."""

        key = event.key()
        if event.key() == Qt.Key.Key_Shift:
            self.shift_down = False
            self.default_cursor = self.standard_cursor
            self.setCursor(self.default_cursor)
        event.accept()

    def wheelEvent(self, event: QWheelEvent):
        """
        Handle a mouse wheel rotation.
        """
        self.editor.wheelEvent(event)

    def resizeEvent(self, event: QResizeEvent = None, updateDiagram: bool = True):
        """
        Widget resized, recompute some state.
        """
        if updateDiagram:
            self.view.set_size_diagram()

        # recalculate the "key" tile stuff
        self.rectify_key_tile()

        if updateDiagram:
            self.view.set_scene_rect_diagram()

    def enterEvent(self, event: QEnterEvent):
        """

        :param event:
        """
        self.setFocus()

    def leaveEvent(self, event: QEvent):
        """
        The mouse is leaving the widget.

        Raise a EVT_PYSLIPQT_POSITION event with positions set to None.
        We do this so user code can clear any mouse position data, for example.
        """

        self.mouse_x = 0
        self.mouse_y = 0

    def dropEvent(self, event: QDropEvent):
        """

        :param event:
        :return:
        """
        super().dropEvent(event)
        self.editor.dropEvent(event)

    def paintEvent(self, event: QPaintEvent):
        """
        Draw the base map and then the layers on top.
        """

        # The "key" tile position is maintained by other code, we just
        # assume it's set.  Figure out how to draw tiles, set up 'row_list' and
        # 'col_list' which are list of tile coords to draw (row and colums).

        col_list = []
        x_coord = self.key_tile_left
        x_pix_start = self.key_tile_x_offset
        while x_pix_start < self.view_width:
            col_list.append(x_coord)
            if x_coord >= self.num_tiles_x - 1:
                break
            x_coord = (x_coord + 1) % self.num_tiles_x
            x_pix_start += self.tile_height

        row_list = []
        y_coord = self.key_tile_top
        y_pix_start = self.key_tile_y_offset
        while y_pix_start < self.view_height:
            row_list.append(y_coord)
            if y_coord >= self.num_tiles_y - 1:
                break
            y_coord = (y_coord + 1) % self.num_tiles_y
            y_pix_start += self.tile_height

        # Ready to update the view
        # prepare the canvas
        painter = QPainter()
        painter.begin(self)

        # paste all background tiles onto the view
        x_pix = self.key_tile_x_offset
        for x in col_list:
            y_pix = self.key_tile_y_offset
            for y in row_list:
                painter.drawPixmap(x_pix, y_pix, self.tile_src.GetTile(x, y))
                y_pix += self.tile_height
            x_pix += self.tile_width

        painter.end()

    def geo_to_view(self, longitude: float, latitude: float) -> Union[None, Tuple[float, float]]:
        """
        Convert a geo coord to view.

        geo  tuple (xgeo, ygeo)

        Return a tuple (xview, yview) in view coordinates.
        Assumes point is in view.
        :param longitude:
        :param latitude:
        :return: x, y
        """

        # convert the Geo position to tile coordinates
        if longitude is not None:
            tx, ty = self.tile_src.Geo2Tile(longitude=longitude, latitude=latitude)

            # using the key_tile_* variables to convert to view coordinates
            xview = (tx - self.key_tile_left) * self.tile_width + self.key_tile_x_offset
            yview = (ty - self.key_tile_top) * self.tile_height + self.key_tile_y_offset

            return xview, yview
        else:
            return None

    # def geo_to_view_masked(self, longitude: float, latitude: float) -> Union[None, Tuple[float, float]]:
    #     """
    #     Convert a geo (lon+lat) position to view pixel coords.
    #     Return a tuple (xview, yview) of point if on-view,or None
    #     if point is off-view.
    #     :param longitude:
    #     :param latitude:
    #     :return: x, y
    #     """
    #
    #     if self.view_llon <= longitude <= self.view_rlon and self.view_blat <= latitude <= self.view_tlat:
    #         return self.geo_to_view(longitude, latitude)
    #
    #     return None

    def view_to_geo_float(self, xview: float, yview: float) -> Tuple[Union[None, float], Union[None, float]]:
        """
        Convert a view coords position to a geo coords position.
        Returns a tuple of geo coords (longitude, latitude) if the cursor is over map
        tiles, else returns None.
        Note: the 'key' tile information must be correct.
        :param xview: x position
        :param yview: y position
        :return: longitude, latitude
        """
        min_lon, max_lon, min_lat, max_lat = self.tile_src.GetExtent()

        x_from_key = xview - self.key_tile_x_offset
        y_from_key = yview - self.key_tile_y_offset

        # get view point as tile coordinates
        xtile: float = self.key_tile_left + x_from_key / self.tile_width
        ytile: float = self.key_tile_top + y_from_key / self.tile_height

        longitude, latitude = self.tile_src.Tile2Geo(xtile, ytile)

        if not (min_lon <= longitude <= max_lon):
            return None, None

        if not (min_lat <= latitude <= max_lat):
            return None, None

        return longitude, latitude

    def view_to_geo(self, xview: float, yview: float) -> Tuple[Union[None, float], Union[None, float]]:
        """
        Convert a view coords position to a geo coords position.
        Returns a tuple of geo coords (longitude, latitude) if the cursor is over map
        tiles, else returns None.
        Note: the 'key' tile information must be correct.
        :param xview: x position
        :param yview: y position
        :return: longitude, latitude
        """
        min_lon, max_lon, min_lat, max_lat = self.tile_src.GetExtent()

        x_from_key = xview - self.key_tile_x_offset
        y_from_key = yview - self.key_tile_y_offset

        # get view point as tile coordinates
        xtile: int = int(self.key_tile_left + x_from_key / self.tile_width)
        ytile: int = int(self.key_tile_top + y_from_key / self.tile_height)

        longitude, latitude = self.tile_src.Tile2Geo(xtile, ytile)

        if not (min_lon <= longitude <= max_lon):
            return None, None

        if not (min_lat <= latitude <= max_lat):
            return None, None

        return longitude, latitude

    ######
    # PEX - Point & EXtension.
    #
    # These functions encapsulate the code that finds the extent of an object.
    # They all return a tuple (point, extent) where 'point' is the placement
    # point of an object (or list of points for a polygon) and an 'extent'
    # tuple (lx, rx, ty, by) [left, right, top, bottom].
    ######

    def pex_point(self, place: Place, xgeo: float, latitude: float, x_off: float, y_off: float, radius: float):
        """Convert point object geo position to point & extent in view coords.

        place         placement string
        geo           point position tuple (xgeo, ygeo)
        x_off, y_off  X and Y offsets
        radius        radius of the point

        Return a tuple of point and extent origins (point, extent) where 'point'
        is (px, py) and extent is (elx, erx, ety, eby) (both in view coords).
        Return None for extent if extent is completely off-view.

        The 'extent' here is the extent of the point+radius.
        """

        # get point view coords
        xview, yview = self.geo_to_view(xgeo, latitude)
        point = self.point_placement(place, xview, yview, x_off, y_off)
        (px, py) = point

        # extent = (left, right, top, bottom) in view coords
        elx = px - radius
        erx = px + radius
        ety = py - radius
        eby = py + radius
        extent = (elx, erx, ety, eby)

        # decide if point extent is off-view
        if erx < 0 or elx > self.view_width or eby < 0 or ety > self.view_height:
            extent = None

        return point, extent

    def pex_point_view(self, place: Place, xview: float, yview: float, x_off: float, y_off: float, radius: float):
        """Convert point object view position to point & extent in view coords.

        place         placement string
        view          point position tuple (xview, yview)
        x_off, y_off  X and Y offsets

        Return a tuple of point and extent origins (point, extent) where 'point'
        is (px, py) and extent is (elx, erx, ety, eby) (both in view coords).
        Return None for point or extent if completely off-view.

        The 'extent' here is the extent of the point+radius.
        """

        # get point view coords and perturb point to placement
        point = self.point_placement_view(place, xview, yview, x_off, y_off)
        (px, py) = point

        # extent = (left, right, top, bottom) in view coords
        elx = px - radius
        erx = px + radius
        ety = py - radius
        eby = py + radius
        extent = (elx, erx, ety, eby)

        # decide if extent is off-view
        if erx < 0 or elx > self.view_width or eby < 0 or ety > self.view_height:
            extent = None

        return point, extent

    def pex_extent(self, place: Place, xgeo: float, latitude: float, x_off: float, y_off: float, w: int, h: int,
                   image=False):
        """Convert object geo position to position & extent in view coords.

        place         placement string
        geo           point position tuple (xgeo, ygeo)
        x_off, y_off  X and Y offsets
        w, h          width and height of extent in pixels
        image         True if we are placing an image.  Required because an image
                      and text extents have DIFFERENT ORIGINS!

        Return a tuple ((px, py), (elx, erx, ety, eby)) of point and extent
        data where '(px, py)' is the point and '(elx, erx, ety, eby)' is the
        extent.  Both point and extent are in view coordinates.

        Return None for point or extent if either is completely off-view.

        An extent object can be either an image object or a text object.
        """

        # get point view coords
        vpoint = self.geo_to_view(xgeo, latitude)
        (vpx, vpy) = vpoint

        # get extent limits
        # must take into account 'place', 'x_off' and 'y_off'
        point = self.extent_placement(place, vpx, vpy, x_off, y_off, w, h, image=image)
        (px, py) = point

        # extent = (left, right, top, bottom) in view coords
        # this is different for images
        elx = px
        erx = px + w
        if image:
            ety = py
            eby = py + h
        else:
            ety = py - h
            eby = py

        extent = (elx, erx, ety, eby)

        # decide if point is off-view
        if vpx < 0 or vpx > self.view_width or vpy < 0 or vpy > self.view_height:
            vpoint = None

        # decide if extent is off-view
        if erx < 0 or elx > self.view_width or eby < 0 or ety > self.view_height:
            # no extent if ALL of extent is off-view
            extent = None

        return vpoint, extent

    def pex_extent_view(self, place, xview, yview, x_off, y_off, w, h, image=False):
        """Convert object view position to point & extent in view coords.

        place         placement string
        view          point position tuple (xview, yview) (view coords)
        x_off, y_off  X and Y offsets
        w, h          width and height of extent in pixels
        image         True if we are placing an image.  Required because an image
                      and text extents have DIFFERENT ORIGINS!

        Return a tuple of point and extent origins (point, extent) where 'point'
        is (px, py) and extent is (elx, erx, ety, eby) (both in view coords).
        Return None for extent if extent is completely off-view.

        Takes size of extent object into consideration.
        """

        # get point view coords and perturb point to placement origin
        # we ignore offsets for the point as they apply to the extent only

        point = self.point_placement_view(place, xview, yview, 0, 0)

        # get extent view coords (ix and iy)
        (px, py) = point
        (ix, iy) = self.extent_placement(place, px, py, x_off, y_off, w, h, image=False)

        # extent = (left, right, top, bottom) in view coords
        # this is different for images
        if image:
            # perturb extent coords to edges of image
            if place == Place.Center:
                elx = px - w / 2
                ety = py - h / 2
            elif place == Place.CenterNorth:
                elx = px - w / 2
                ety = py + y_off
            elif place == Place.NorthEast:
                elx = px - w - x_off
                ety = py + y_off
            elif place == Place.CenterEast:
                elx = px - w - x_off
                ety = py - h / 2
            elif place == Place.SouthEast:
                elx = px - w - x_off
                ety = py - h - y_off
            elif place == Place.CenterSouth:
                elx = px - w / 2
                ety = py - h - y_off
            elif place == Place.SouthWest:
                elx = px + x_off
                ety = py - h - y_off
            elif place == Place.CenterWest:
                elx = px + x_off
                ety = py - h / 2
            elif place == Place.NorthWest:
                elx = px + x_off
                ety = py + y_off
            else:
                raise Exception('Unsupported place: ' + place)

            erx = elx + w
            eby = ety + h
        else:
            elx = ix
            erx = ix + w
            ety = iy - h
            eby = iy

        extent = (elx, erx, ety, eby)

        # decide if point is off-view
        if px < 0 or px > self.view_width or py < 0 or py > self.view_height:
            point = None

        # decide if extent is off-view
        if erx < 0 or elx > self.view_width or eby < 0 or ety > self.view_height:
            extent = None

        return point, extent

    def pex_polygon(self, place: Place, poly: List[Tuple[float, float]], x_off: float, y_off: float):
        """Convert polygon/line obj geo position to points & extent in view coords.

        place         placement
        poly          list of point position tuples (xgeo, ygeo)
        x_off, y_off  X and Y offsets

        Return a tuple of point and extent (point, extent) where 'point' is a
        list of (px, py) and extent is (elx, erx, ety, eby) (both in view coords).
        Return None for extent if extent is completely off-view.
        """

        # get polygon/line points in perturbed view coordinates
        view_points = []
        for lon, lat in poly:
            xview, yview = self.geo_to_view(longitude=lon, latitude=lat)
            point = self.point_placement(place, xview, yview, x_off, y_off)
            view_points.append(point)

        # extent = (left, right, top, bottom) in view coords
        elx = min(view_points, key=lambda x: x[0])[0]
        erx = max(view_points, key=lambda x: x[0])[0]
        ety = min(view_points, key=lambda x: x[1])[1]
        eby = max(view_points, key=lambda x: x[1])[1]
        extent = (elx, erx, ety, eby)

        # decide if extent is off-view
        res_ex = None  # assume extent is off-view
        for px, py in view_points:
            if (0 <= px < self.view_width) and (0 <= py < self.view_height):
                res_ex = extent  # at least some of extent is on-view
                break

        return view_points, res_ex

    def pex_polygon_view(self, place: Place, poly: List[Tuple[float, float]], x_off: float, y_off: float):
        """Convert polygon/line obj view position to points & extent in view coords.

        place         placement string
        poly          list of point position tuples (xview, yview)
        x_off, y_off  X and Y offsets

        Return a tuple of point and extent origins (point, extent) where 'point'
        is a list of (px, py) and extent is (elx, erx, ety, eby) (both in view
        coords).  Return None for extent if extent is completely off-view.
        """

        # get polygon/line points in view coordinates
        view = []
        for (xview, yview) in poly:
            point = self.point_placement_view(place, xview, yview, x_off, y_off)
            view.append(point)

        # get extent - max/min x and y
        # extent = (left, right, top, bottom) in view coords
        elx = min(view, key=lambda x: x[0])[0]
        erx = max(view, key=lambda x: x[0])[0]
        ety = min(view, key=lambda x: x[1])[1]
        eby = max(view, key=lambda x: x[1])[1]
        extent = (elx, erx, ety, eby)

        # decide if polygon/line or extent are off-view
        res_ex = None
        for (px, py) in view:
            if (0 <= px < self.view_width) and (0 <= py < self.view_height):
                res_ex = extent
                break

        return view, res_ex

    ######
    # Placement routines instead of original 'exec' code.
    # Code in test_assumptions.py shows this is faster.
    ######

    @staticmethod
    def point_placement(place: Place, x: float, y: float, x_off: float, y_off: float):
        """Perform map-relative placement for a single point.

        place         placement key string
        x, y          point view coordinates
        x_off, y_off  the X and Y offset values

        Returns a tuple (x, y) in view coordinates.
        """

        # adjust the X, Y coordinates relative to the origin
        if place == Place.Center:
            pass
        elif place == Place.NorthWest:
            x += x_off
            y += y_off
        elif place == Place.CenterNorth:
            y += y_off
        elif place == Place.NorthEast:
            x += -x_off
            y += y_off
        elif place == Place.CenterEast:
            x += -x_off
        elif place == Place.SouthEast:
            x += -x_off
            y += -y_off
        elif place == Place.CenterSouth:
            y += -y_off
        elif place == Place.SouthWest:
            x += x_off
            y += -y_off
        elif place == Place.CenterWest:
            x += x_off
        else:
            raise Exception('Unsupported place: ' + place.value)

        return x, y

    def point_placement_view(self, place: Place, x: float, y: float, x_off: float, y_off: float):
        """Perform view-relative placement for a single point.

        place         placement key string
        x, y          point view coordinates
        x_off, y_off  the X and Y offset values

        Returns a tuple (x, y) in view coordinates.
        """

        dcw = self.view_width
        dch = self.view_height
        dcw2 = dcw / 2
        dch2 = dch / 2

        # adjust the X, Y coordinates relative to the origin
        # offsets are always away from the nearest edge
        if place == Place.Center:
            x += dcw2
            y += dch2
        elif place == Place.NorthWest:
            x += x_off
            y += y_off
        elif place == Place.CenterNorth:
            x += dcw2
            y += y_off
        elif place == Place.NorthEast:
            x += dcw - x_off
            y += y_off
        elif place == Place.CenterEast:
            x += dcw - x_off
            y += dch2
        elif place == Place.SouthEast:
            x += dcw - x_off
            y += dch - y_off
        elif place == Place.CenterSouth:
            x += dcw2
            y += dch - y_off
        elif place == Place.SouthWest:
            x += x_off
            y += dch - y_off
        elif place == Place.CenterWest:
            x += x_off
            y += dch2
        else:
            raise Exception('Unsupported place: ' + place.value)

        return x, y

    def easy_lat_lon_to_x_y(self, lat: float, lon: float) -> Tuple[float, float]:
        """

        :param lat:
        :param lon:
        :return:
        """
        point, extent = self.pex_point(place=Place.Center,
                                       xgeo=lat,
                                       latitude=lon,
                                       x_off=0.0,
                                       y_off=0.0,
                                       radius=1)
        return point[0], point[1]

    @staticmethod
    def extent_placement(place: Place, x: float, y: float, x_off: float, y_off: float, w: int, h: int,
                         image: bool = False):
        """Perform map-relative placement of an extent.

        place         placement key string
        x, y          view coords of point
        x_off, y_off  offset from point (pixels)
        w, h          width, height of the extent (pixels)
        image         True if we are placing an image.  Required because an image
                      and text extents have DIFFERENT ORIGINS!

        Returns a tuple (x, y).
        """

        w2 = w / 2
        h2 = h / 2

        if image:
            if place == Place.Center:
                x += -w2
                y += -h2
            elif place == Place.NorthWest:
                x += x_off
                y += y_off
            elif place == Place.CenterNorth:
                x += -w2
                y += y_off
            elif place == Place.NorthEast:
                x += -x_off - w
                y += y_off
            elif place == Place.CenterEast:
                x += -x_off - w
                y += -h2
            elif place == Place.SouthEast:
                x += -x_off - w
                y += -y_off - h
            elif place == Place.CenterSouth:
                x += -w2
                y += -y_off - h
            elif place == Place.SouthWest:
                x += x_off
                y += -y_off - h
            elif place == Place.CenterWest:
                x += x_off
                y += -h2
            else:
                raise Exception('Unsupported place: ' + place.value)
        else:
            if place == Place.Center:
                x += -w2
                y += h2
            elif place == Place.NorthWest:
                x += x_off
                y += y_off + h
            elif place == Place.CenterNorth:
                x += -w2
                y += y_off + h
            elif place == Place.NorthEast:
                x += -x_off - w
                y += y_off + h
            elif place == Place.CenterEast:
                x += -x_off - w
                y += h2
            elif place == Place.SouthEast:
                x += -x_off - w
                y += -y_off
            elif place == Place.CenterSouth:
                x += -w2
                y += -y_off
            elif place == Place.SouthWest:
                x += x_off
                y += -y_off
            elif place == Place.CenterWest:
                x += x_off
                y += h2
            else:
                raise Exception('Unsupported place: ' + place.value)

        return x, y

    def set_zoom_level(self,
                       level: int,
                       view_x: float | None = None,
                       view_y: float | None = None) -> bool:
        """
        Zoom to a map level.

        :param level:  map level to zoom to
        :param view_x: view x coordinate
        :param view_y: view y coordinate

        Change the map zoom level to that given. Returns True if the zoom
        succeeded, else False. If False is returned the method call has no effect.
        Same operation as .GotoLevel() except we try to maintain the geo position
        under the cursor.
        """

        # if not given cursor coords, assume view centre
        if view_x is None:
            view_x = self.view_width // 4

        if view_y is None:
            view_y = self.view_height // 4

        # get geo coords of view point
        longitude, latitude = self.view_to_geo_float(view_x, view_y)

        # get tile source to use the new level
        result = self.tile_src.set_level(level)

        if result:
            # zoom worked, adjust state variables
            self.level = level

            # move to new level
            self.tile_src.GetInfo(level)

            # finally, pan to original map centre (updates widget)
            self.pan_position(longitude, latitude, view_x, view_y)

            # to set some state variables
            self.resizeEvent()

        self.zoom_callback(level)

        return result

    def pan_position(self, longitude: float, latitude: float, view_x: int = None, view_y: int = None):
        """
        Pan the given geo position in the current map zoom level.

        We just adjust the key tile to place the required geo position at the
        given view coordinates.  If that is not possible, just centre in either
        the X or Y directions, or both.

        :param longitude:
        :param latitude:
        :param view_x:
        :param view_y:
        :return:
        """

        # if not given a "view", assume the view centre coordinates
        if view_x is None:
            view_x = self.view_width // 2

        if view_y is None:
            view_y = self.view_height // 2

        if longitude is None:
            return

        # convert the geo position to a tile position
        tile_x, tile_y = self.tile_src.Geo2Tile(longitude, latitude)

        # determine what the new key tile should be
        # figure out number of tiles from centre point to edges
        tx = view_x / self.tile_width
        ty = view_y / self.tile_height

        # calculate tile coordinates of the top-left corner of the view
        key_tx = tile_x - tx
        key_ty = tile_y - ty

        (key_tile_left, x_offset) = divmod(key_tx, 1)
        self.key_tile_left = int(key_tile_left)
        self.key_tile_x_offset = -int(x_offset * self.tile_width)

        (key_tile_top, y_offset) = divmod(key_ty, 1)
        self.key_tile_top = int(key_tile_top)
        self.key_tile_y_offset = -int(y_offset * self.tile_height)

        # adjust key tile, if necessary
        self.rectify_key_tile()

        # redraw the widget
        self.update()

    def rectify_key_tile(self) -> None:
        """
        Adjust state variables to ensure map centred if map is smaller than view.
        Otherwise don't allow edges to be exposed.
        Adjusts the "key" tile variables to ensure proper presentation.
        Relies on .map_width, .map_height and .key_tile_* being set.
        """

        # check map in X direction
        if self.map_width < self.view_width:
            # map < view, fits totally in view, centre in X
            self.key_tile_left = 0
            self.key_tile_x_offset = (self.view_width - self.map_width) // 2
        else:
            # if key tile out of map in X direction, rectify
            if self.key_tile_left < 0:
                self.key_tile_left = 0
                self.key_tile_x_offset = 0
            else:
                # if map left/right edges showing, cover them
                show_len = (self.num_tiles_x - self.key_tile_left) * self.tile_width + self.key_tile_x_offset
                if show_len < self.view_width:
                    # figure out key tile X to have right edge of map and view equal
                    tiles_showing = self.view_width / self.tile_width
                    int_tiles = int(tiles_showing)
                    self.key_tile_left = self.num_tiles_x - int_tiles - 1
                    self.key_tile_x_offset = -int((1.0 - (tiles_showing - int_tiles)) * self.tile_width)

        # now check map in Y direction
        if self.map_height < self.view_height:
            # map < view, fits totally in view, centre in Y
            self.key_tile_top = 0
            self.key_tile_y_offset = (self.view_height - self.map_height) // 2
        else:
            if self.key_tile_top < 0:
                # map top edge showing, cover
                self.key_tile_top = 0
                self.key_tile_y_offset = 0
            else:
                # if map bottom edge showing, cover
                show_len = (self.num_tiles_y - self.key_tile_top) * self.tile_height + self.key_tile_y_offset
                if show_len < self.view_height:
                    # figure out key tile Y to have bottom edge of map and view equal
                    tiles_showing = self.view_height / self.tile_height
                    int_tiles = int(tiles_showing)
                    self.key_tile_top = self.num_tiles_y - int_tiles - 1
                    self.key_tile_y_offset = -int((1.0 - (tiles_showing - int_tiles)) * self.tile_height)

    def zoom_level_position(self, level: int, longitude: float, latitude: float):
        """Zoom to a map level and pan to the given position in the map.

        level  map level to zoom to
        posn  a tuple (xgeo, ygeo)
        """

        if self.set_zoom_level(level):
            self.pan_position(longitude, latitude)

    def get_level_and_position(self, place=Place.Center):
        """Get the level and geo position of a cardinal point within the view.

        place  a placement string specifying the point in the view
               for which we require the geo position

        Returns a tuple (level, geo) where 'geo' is (geo_x, geo_y).
        """

        view_x, view_y = self.point_placement_view(place, 0, 0, 0, 0, )
        longitude, latitude = self.view_to_geo(view_x, view_y)

        return self.level, longitude, latitude

    def set_key_from_centre(self, longitude: float, latitude: float):
        """Set 'key' tile stuff from given geo at view centre.

        geo  geo coords of centre of view

        We need to assume little about which state variables are set.
        Only assume these are set:
            self.tile_width
            self.tile_height
        """
        if longitude is None:
            return

        ctile_tx, ctile_ty = self.tile_src.Geo2Tile(longitude, latitude)

        int_ctile_tx = int(ctile_tx)
        int_ctile_ty = int(ctile_ty)

        frac_ctile_tx = ctile_tx - int_ctile_tx
        frac_ctile_ty = ctile_ty - int_ctile_ty

        ctile_xoff = self.view_width // 2 - self.tile_width * frac_ctile_tx
        ctile_yoff = self.view_height // 2 - self.tile_height * frac_ctile_ty

        num_whole_x = ctile_xoff // self.tile_width
        num_whole_y = ctile_yoff // self.tile_height

        xmargin = ctile_xoff - num_whole_x * self.tile_width
        ymargin = ctile_yoff - num_whole_y * self.tile_height

        # update the 'key' tile state variables
        self.key_tile_left = int_ctile_tx - num_whole_x - 1
        self.key_tile_top = int_ctile_ty - num_whole_y - 1
        self.key_tile_x_offset = self.tile_width - xmargin
        self.key_tile_y_offset = self.tile_height - ymargin

        # centre map in view if map < view
        if self.key_tile_left < 0:
            self.key_tile_left = 0
            self.key_tile_x_offset = (self.view_width - self.map_width) // 2

        if self.key_tile_top < 0:
            self.key_tile_top = 0
            self.key_tile_y_offset = (self.view_height - self.map_height) // 2

    @staticmethod
    def colour_to_internal(colour: Union[str, QColor, Tuple[int, int, int, int]]):
        """
        Convert a colour in one of various forms to an internal format.

        colour  either a HEX string ('#RRGGBBAA')
                or a tuple (r, g, b, a)
                or a colour name ('red')

        Returns internal form:  (r, g, b, a)
        """

        if isinstance(colour, str):
            # expect '#RRGGBBAA' form
            if len(colour) != 9 or colour[0] != '#':
                # assume it's a colour *name*
                # we should do more checking of the name here, though it looks
                # like PySide6 defaults to a colour if the name isn't recognized
                c = QColor(colour)
                result = (c.red(), c.blue(), c.green(), c.alpha())
            else:
                # we try for a colour like '#RRGGBBAA'
                r = int(colour[1:3], 16)
                g = int(colour[3:5], 16)
                b = int(colour[5:7], 16)
                a = int(colour[7:9], 16)
                result = (r, g, b, a)
        elif isinstance(colour, QColor):
            # if it's a QColor, get float RGBA values, convert to ints
            result = [int(v * 255) for v in tuple(colour.getRgbF())]
        else:

            # we assume a list or tuple
            try:
                len_colour = len(colour)
            except TypeError:
                msg = ("Colour value '%s' is not in the form '(r, g, b, a)'"
                       % str(colour))
                raise Exception(msg)

            if len_colour != 4:
                msg = ("Colour value '%s' is not in the form '(r, g, b, a)'"
                       % str(colour))
                raise Exception(msg)
            result = []
            for v in colour:
                try:
                    v = int(v)
                except ValueError:
                    msg = ("Colour value '%s' is not in the form '(r, g, b, a)'"
                           % str(colour))
                    raise Exception(msg)
                if v < 0 or v > 255:
                    msg = ("Colour value '%s' is not in the form '(r, g, b, a)'"
                           % str(colour))
                    raise Exception(msg)
                result.append(v)
            result = tuple(result)

        return result

    ################################################################################
    # Below are the "external" API methods.
    ################################################################################

    ######
    # "add a layer" routines
    ######

    def GotoLevel(self, level: int):
        """
        Use a new tile level.
        :param: level  the new tile level to use.
        Returns True if all went well.
        """

        if not self.tile_src.set_level(level):
            return False  # couldn't change level

        self.level = level

        self.tile_src.GetInfo(level)

        # to set some state variables
        self.resizeEvent(updateDiagram=False)

        return True

    def go_to_position(self, longitude: float, latitude: float):
        """Set view to centre on a geo position in the current level.

        geo  a tuple (xgeo,ygeo) to centre view on

        Recalculates the key tile info.
        """
        if longitude is None:
            return

        # get fractional tile coords of required centre of view
        x_tile, y_tile = self.tile_src.Geo2Tile(longitude=longitude, latitude=latitude)

        # get view size in half widths and height
        w2 = self.view_width / 2
        h2 = self.view_height / 2

        # get tile coords of view left and top edges
        view_tile_x = x_tile - (w2 / self.tile_width)
        view_tile_y = y_tile - (h2 / self.tile_height)

        # calculate the key tile coords and offsets
        key_tile_x = int(view_tile_x)
        key_tile_y = int(view_tile_y)

        key_offset_x = - int((view_tile_x - key_tile_x) * self.tile_width)
        key_offset_y = - int((view_tile_y - key_tile_y) * self.tile_height)

        # update the key tile info
        self.key_tile_left = key_tile_x
        self.key_tile_top = key_tile_y
        self.key_tile_x_offset = key_offset_x
        self.key_tile_y_offset = key_offset_y

        # centre map in view if map < view
        if self.key_tile_left < 0 or self.key_tile_x_offset > 0:
            self.key_tile_left = 0
            self.key_tile_x_offset = (self.view_width - self.map_width) // 2

        if self.key_tile_top < 0 or self.key_tile_y_offset > 0:
            self.key_tile_top = 0
            self.key_tile_y_offset = (self.view_height - self.map_height) // 2

        # redraw the display
        self.update()

        self.position_callback(latitude, longitude, x_tile, y_tile)

    def go_to_level_and_position(self, level: int, longitude: float, latitude: float):
        """
        Goto a map level and set view to centre on a position.

        :param level: zoom level (int value)
        :param longitude: longitude (deg)
        :param latitude: latitude (deg)
        """

        if self.GotoLevel(level):
            self.go_to_position(longitude=longitude, latitude=latitude)

    def change_tile_set(self, tile_src: Tiles) -> Tiles:
        """
        Change the source of tiles.

        :param tile_src:  the new Tiles object to use

        Returns the previous Tiles object, None if none.

        Refreshes the display and tries to maintain the same position
        and zoom level.  May change the zoom level if the current level doesn't
        exist in the new Tiles.
        """

        # get level and geo position of view centre
        level, longitude, latitude = self.get_level_and_position()
        # print('level=%s, geo=(%s, %s)' % (str(level), str(longitude), str(latitude)))

        # remember old Tiles object
        old_tile_set = self.tile_src

        # get levels in new tile set and see if we can display at the current level
        new_levels = tile_src.levels
        new_max_level = tile_src.max_level
        new_min_level = tile_src.min_level

        if level > new_max_level:
            level = new_max_level

        if level < new_min_level:
            level = new_min_level

        # set new tile source and set some state
        self.tile_src = tile_src

        self.level = level

        # set callback from Tile source object when tile(s) available
        self.tile_src.setCallback(self.on_tile_available)

        # set the new zoom level to the old
        if not tile_src.set_level(self.level):
            # can't use old level, make sensible choice
            if self.level < self.min_level:
                self.level = self.min_level

            elif self.level > self.max_level:
                self.level = self.max_level

            # if we can't change level now, raise an error exception
            if not tile_src.set_level(self.level):
                raise Exception('Trying to use level %s in tile obj %s, '
                                'levels available are %s'
                                % (str(self.level),
                                   str(tile_src), str(tile_src.levels)))

        # TODO: MUST SET KEY TILE STUFF HERE
        self.set_key_from_centre(longitude, latitude)

        # back to old level+centre, and refresh the display
        self.zoom_level_position(level, longitude, latitude)

        return old_tile_set

    def get_selected(self) -> List[QGraphicsItem]:
        """
        Get the selection
        :return:
        """
        return self.view.get_selected()



# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A server Tiles object for pySlipQt tiles.

All server tile sources should inherit from this class.
For example, see osm_tiles.py.
"""
import http.client
import os
import time
import math
import traceback
import urllib
from urllib import request
from urllib.error import HTTPError
import queue
from typing import List, Union
from collections.abc import Callable
from warnings import warn
from PySide6.QtGui import QPixmap, QColor

from VeraGrid.Gui.Diagrams.MapWidget.Tiles.base_tiles import BaseTiles
from VeraGrid.Gui.Diagrams.MapWidget.Tiles.tile_worker import TileWorker


# # set how old disk-cache tiles can be before we re-request them from the
# # server.  this is the number of days old a tile is before we re-request.
# # if 'None', never re-request tiles after first satisfied request.
# RefreshTilesAfterDays = 60


class Tiles(BaseTiles):
    """
    A tile object to source server tiles for the widget.
    """

    def __init__(self,
                 tile_set_name: str,
                 tile_set_short_name: str,
                 tile_set_version: str,
                 levels: List[int],
                 tile_width: int,
                 tile_height: int,
                 tiles_dir: str,
                 max_lru: int,
                 servers: List[str],
                 url_path: str,
                 max_server_requests: int,
                 http_proxy,
                 re_fetch_days: int = 60,
                 attribution: str = ""):
        """
        Initialise a Tiles instance.
        :param tile_set_name: Name of the tile set.
        :param tile_set_short_name: Short name of the tile set.
        :param tile_set_version: Version of the tile set.
        :param levels: a list of level numbers that are to be served
        :param tile_width: width of each tile in pixels
        :param tile_height: height of each tile in pixels
        :param tiles_dir: path to on-disk tile cache directory
        :param max_lru: maximum number of tiles cached in-memory
        :param servers: list of tile servers
        :param url_path: path on server to each tile
        :param max_server_requests: maximum number of requests per server
        :param http_proxy: proxy to use if required
        :param re_fetch_days: fetch new server tile if older than this in days (0 means don't ever update tiles)
        """
        # perform the base class initialization
        super().__init__(levels, tile_width, tile_height, tiles_dir, max_lru)

        # allowed file types and associated values
        self.AllowedFileTypes = {'png': 'PNG', 'jpg': 'JPG'}

        # the number of seconds in a day
        self.SecondsInADay = 60 * 60 * 24

        self.tile_set_name = tile_set_name
        self.tile_set_short_name = tile_set_short_name
        self.tile_set_version = tile_set_version

        self.attribution_string = attribution

        # prepare the tile cache directory, if required
        # we have to do this *before* the base class initialization!
        for level in levels:
            level_dir = os.path.join(tiles_dir, '%d' % level)
            if not os.path.isdir(level_dir):
                os.makedirs(level_dir)

        # save params not saved in super()
        self.servers = servers
        self.url_path = url_path
        self.max_requests = max_server_requests
        self.http_proxy = http_proxy
        self.refresh_tiles_after_days = re_fetch_days

        # callback must be set by higher-level code
        self.callback: Union[None, Callable[[int, float, float, QPixmap, bool], None]] = None

        # calculate a re-request age, if specified
        self.re_request_age = (time.time() - self.refresh_tiles_after_days * self.SecondsInADay)

        # tiles extent for tile data (left, right, top, bottom)
        self.extent = (-180.0, 180.0, -85.0511, 85.0511)

        self.level = levels[len(levels) - 1]
        self.num_tiles_x = 0
        self.num_tiles_y = 0
        self.ppd_x = 0
        self.ppd_y = 0

        # figure out tile filename extension from 'url_path'
        tile_extension = os.path.splitext(url_path)[1][1:]
        tile_extension_lower = tile_extension.lower()  # ensure lower case

        if tile_extension_lower == "":
            tile_extension_lower = 'jpg'

        # determine the file bitmap type
        try:
            self.filetype = self.AllowedFileTypes[tile_extension_lower]
        except KeyError:
            raise TypeError("Bad tile_extension value, got '%s', "
                            "expected one of %s"
                            % (str(tile_extension),
                               str(self.AllowedFileTypes.keys()))) from None

        # compose the expected 'Content-Type' string on request result
        # if we get here we know the extension is in self.AllowedFileTypes
        # if tile_extension_lower in ['png', 'jpg']:

        self.content_type = 'image/png'

        # set the list of queued unsatisfied requests to 'empty'
        self.queued_requests = {}

        # prepare the "pending" and "error" images
        self.pending_tile = QPixmap(256, 256)
        self.pending_tile.fill(QColor.fromRgb(50, 50, 50, 255))

        self.error_tile = QPixmap(256, 256)
        self.error_tile.fill(QColor.fromRgb(255, 0, 0, 255))

        # define the error messages for various failures
        StatusError = {401: 'Looks like you need to be authorised for this server.',
                       404: 'You might need to check the tile addressing for this server.',
                       429: 'You are asking for too many tiles.', }

        # test for firewall - use proxy (if supplied)
        test_url = self.servers[0] + self.url_path.format(Z=0, X=0, Y=0)
        try:
            r = request.Request(test_url, headers={'User-Agent': 'VeraGrid 5'})
            response = request.urlopen(r).read()

        except HTTPError as e:
            # if it's fatal, log it and die, otherwise try a proxy
            status_code = e.code
            warn('Error: test_url=%s, status_code=%s' % (test_url, str(status_code)))
            error_msg = StatusError.get(status_code, None)
            if status_code:
                msg = "\nYou got a " + str(status_code) + " (" + str(error_msg) + ") error from: " + str(test_url)
                print(msg)
                # raise RuntimeError(msg) from None

            warn('%s exception doing simple connection to: %s' % (type(e).__name__, test_url))
            warn(''.join(traceback.format_exc()))

            if http_proxy:
                proxy = request.ProxyHandler({'http': http_proxy})
                opener = request.build_opener(proxy)
                request.install_opener(opener)
                try:
                    request.urlopen(test_url)
                except:
                    msg = "Using HTTP proxy but still can't get through a firewall!"
                    print(msg)
                    # raise Exception(msg) from None
            else:
                msg = "There is a firewall but you didn't give me an HTTP proxy to get through it?"
                print(msg)
                # raise Exception(msg) from None
        except urllib.error.URLError as e:
            print(e)

        except http.client.IncompleteRead as e:
            print(e)

        # set up the request queue and worker threads
        self.request_queue = queue.Queue()  # entries are (level, x, y)
        self.workers = []
        for server in self.servers:
            for num_thread in range(self.max_requests):
                worker = TileWorker(id_num=num_thread,
                                    server=server,
                                    tile_path=self.url_path,
                                    requests_cue=self.request_queue,
                                    callback=self.tile_is_available,
                                    error_tile=self.error_tile,
                                    content_type=self.content_type,
                                    re_request_age=self.re_request_age,
                                    error_image=self.error_tile,
                                    refresh_tiles_after_days=60)
                self.workers.append(worker)
                worker.start()

    def set_level(self, level: int):
        """
        Prepare to serve tiles from the required level.
        :param level: the required level
        :return: True if level change occurred, else False if not possible.
        """
        # first, CAN we zoom to this level?
        if self.level_in_range(level):

            # get tile info
            info = self.GetInfo(level)
            if info is None:
                return False

            # OK, save new level
            self.level = level
            self.num_tiles_x, self.num_tiles_y, self.ppd_x, self.ppd_y = info

            # flush any outstanding requests.
            # we do this to speed up multiple-level zooms so the user doesn't
            # sit waiting for tiles to arrive that won't be shown.
            self.FlushRequests()

            return True

        else:
            # the zoom is out of bounds...
            return False

    def GetTile(self, x: float, y: float) -> QPixmap:
        """
        Get bitmap for tile at tile coords (x, y) and current level.

        :param x:  X coord of tile required (tile coordinates)
        :param y:  Y coord of tile required (tile coordinates)

        Returns bitmap object for the tile image.
        Tile coordinates are measured from map top-left.

        We override the existing GetTile() method to add code to retrieve
        tiles from the servers if not in on-disk cache.

        We also check the date on the tile from disk-cache.  If "too old",
        return old tile after starting the process to get new tile from servers.
        """

        try:
            # get tile from cache
            tile = self.cache[(self.level, x, y)]
            if self.tile_on_disk(level=self.level, x=x, y=y):
                tile_date = self.cache.tile_date((self.level, x, y))
                if self.re_request_age and (tile_date < self.re_request_age):
                    self.get_server_tile(level=self.level, x=x, y=y)
        except KeyError:
            # not cached, start process of getting tile from 'net, return 'pending' image
            self.get_server_tile(level=self.level, x=x, y=y)
            tile = self.pending_tile

        return tile

    def GetInfo(self, level: int):
        """
        Get tile info for a particular level.

        level  the level to get tile info for

        Returns (num_tiles_x, num_tiles_y, ppd_x, ppd_y) or None if 'level'
        doesn't exist.

        Note that ppd_? may be meaningless for some tiles, so its
        value will be None.

        This method is for server tiles.  It will be overridden for GMT tiles.
        """

        # is required level available?
        if self.level_in_range(level):

            # otherwise get the information
            self.num_tiles_x = int(math.pow(2, level))
            self.num_tiles_y = int(math.pow(2, level))

            return self.num_tiles_x, self.num_tiles_y, None, None
        else:
            return None

    def FlushRequests(self):
        """
        Delete any outstanding tile requests.
        """

        # if we are serving server tiles ...
        if self.servers:
            with self.request_queue.mutex:
                self.request_queue.queue.clear()
            self.queued_requests.clear()

    def get_server_tile(self, level: int, x: float, y: float) -> None:
        """
        Start the process to get a server tile.

        level, x, y  identify the required tile

        If we don't already have this tile (or getting it), queue a request and
        also put the request into a 'queued request' dictionary.  We
        do this since we can't peek into a queue to see what's there.
        """

        tile_key = (level, x, y)
        if tile_key not in self.queued_requests:
            # add tile request to the server request queue
            self.request_queue.put(tile_key)
            self.queued_requests[tile_key] = True

    def tile_on_disk(self, level: int, x: float, y: float):
        """
        Return True if tile at (level, x, y) is on-disk.
        """

        tile_path = self.cache.tile_path((level, x, y))
        return os.path.exists(tile_path)

    def setCallback(self, callback: Callable[[int, float, float, QPixmap, bool], None]):
        """Set the "tile available" callback.

        callback  reference to object to call when tile is found.
        """

        self.callback = callback

    def tile_is_available(self, level: int, x: float, y: float, image: QPixmap, error: bool):
        """
        Callback routine - a 'net tile is available.

        level   level for the tile
        x       x coordinate of tile
        y       y coordinate of tile
        image   tile image data
        error   True if image is 'error' image, don't cache in that case
        """

        # put image into in-memory cache, but error images don't go to disk
        self.cache[(level, x, y)] = image
        if not error:
            self.cache.add(key=(level, x, y), image=image)

        # delete the request from the queued requests
        # note that it may not be there - a level change can flush the dict
        try:
            del self.queued_requests[(level, x, y)]
        except KeyError:
            pass

        # tell the world a new tile is available
        if self.callback:
            self.callback(level, x, y, image, True)
        else:
            msg = f'tile_is_available: self.callback is NOT SET!'
            raise RuntimeError(msg) from None

    def SetAgeThresholdDays(self, num_days):
        """
        Set the tile re-fetch threshold time.

        num_days  number of days before re-fetching tiles

        If 'num_days' is 0 re-fetching is inhibited.
        """

        # update the global in case we instantiate again

        self.refresh_tiles_after_days = num_days

        # recalculate this instance's age threshold in UNIX time
        self.re_request_age = time.time() - self.refresh_tiles_after_days * self.SecondsInADay



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'profiles_from_models_gui.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QAbstractScrollArea, QApplication, QCheckBox,
    QDialog, QFrame, QHBoxLayout, QHeaderView,
    QPushButton, QSizePolicy, QSpacerItem, QTableView,
    QVBoxLayout, QWidget)
from .icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(713, 522)
        self.verticalLayout_2 = QVBoxLayout(Dialog)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.frame_3 = QFrame(Dialog)
        self.frame_3.setObjectName(u"frame_3")
        self.frame_3.setMaximumSize(QSize(16777215, 34))
        self.frame_3.setFrameShape(QFrame.NoFrame)
        self.frame_3.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_6 = QHBoxLayout(self.frame_3)
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.horizontalLayout_6.setContentsMargins(0, 0, 0, 0)
        self.addModelsButton = QPushButton(self.frame_3)
        self.addModelsButton.setObjectName(u"addModelsButton")
        icon = QIcon()
        icon.addFile(u":/Icons/icons/plus.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.addModelsButton.setIcon(icon)

        self.horizontalLayout_6.addWidget(self.addModelsButton)

        self.deleteModelsButton = QPushButton(self.frame_3)
        self.deleteModelsButton.setObjectName(u"deleteModelsButton")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/delete3.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.deleteModelsButton.setIcon(icon1)

        self.horizontalLayout_6.addWidget(self.deleteModelsButton)

        self.horizontalSpacer_5 = QSpacerItem(725, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_5)


        self.verticalLayout_2.addWidget(self.frame_3)

        self.frame_5 = QFrame(Dialog)
        self.frame_5.setObjectName(u"frame_5")
        self.frame_5.setFrameShape(QFrame.NoFrame)
        self.frame_5.setFrameShadow(QFrame.Raised)
        self.verticalLayout = QVBoxLayout(self.frame_5)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, -1, 0, -1)
        self.modelsTableView = QTableView(self.frame_5)
        self.modelsTableView.setObjectName(u"modelsTableView")
        self.modelsTableView.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        self.modelsTableView.setDragDropMode(QAbstractItemView.InternalMove)
        self.modelsTableView.setDefaultDropAction(Qt.MoveAction)
        self.modelsTableView.setAlternatingRowColors(True)
        self.modelsTableView.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.verticalLayout.addWidget(self.modelsTableView)


        self.verticalLayout_2.addWidget(self.frame_5)

        self.frame_7 = QFrame(Dialog)
        self.frame_7.setObjectName(u"frame_7")
        self.frame_7.setMaximumSize(QSize(16777215, 40))
        self.frame_7.setFrameShape(QFrame.NoFrame)
        self.frame_7.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_8 = QHBoxLayout(self.frame_7)
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.horizontalLayout_8.setContentsMargins(0, 0, 0, 0)
        self.matchUsingCodeCheckBox = QCheckBox(self.frame_7)
        self.matchUsingCodeCheckBox.setObjectName(u"matchUsingCodeCheckBox")

        self.horizontalLayout_8.addWidget(self.matchUsingCodeCheckBox)

        self.horizontalSpacer_6 = QSpacerItem(814, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer_6)

        self.acceptModelsButton = QPushButton(self.frame_7)
        self.acceptModelsButton.setObjectName(u"acceptModelsButton")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/gear.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.acceptModelsButton.setIcon(icon2)

        self.horizontalLayout_8.addWidget(self.acceptModelsButton)


        self.verticalLayout_2.addWidget(self.frame_7)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
        self.addModelsButton.setText(QCoreApplication.translate("Dialog", u"Add", None))
        self.deleteModelsButton.setText(QCoreApplication.translate("Dialog", u"Clear", None))
#if QT_CONFIG(tooltip)
        self.matchUsingCodeCheckBox.setToolTip(QCoreApplication.translate("Dialog", u"If checked, the objects are match using the code property, otherwise the idtag property is used", None))
#endif // QT_CONFIG(tooltip)
        self.matchUsingCodeCheckBox.setText(QCoreApplication.translate("Dialog", u"Match using code", None))
        self.acceptModelsButton.setText(QCoreApplication.translate("Dialog", u"Accept", None))
    # retranslateUi




# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
An extended dictionary offering limited LRU entries in the dictionary
and an interface to an unlimited backing store.

https://github.com/rzzzwilson/pyCacheBack
"""


class PyCacheBack(dict):
    """An LRU limited in-memory store fronting an unlimited on-disk store."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # default maximum number of key/value pairs for pyCacheBack
        self.DefaultMaxLRU = 1000

        # default path to tiles directory
        self.DefaultTilesDir = 'tiles'

        self._lru_list = []
        self._max_lru = kwargs.pop('max_lru', self.DefaultMaxLRU)
        self._tiles_dir = kwargs.pop('tiles_dir', self.DefaultTilesDir)

    def __getitem__(self, key):
        if key in self:
            value = super().__getitem__(key)
        else:
            value = self._get_from_back(key)
        self._reorder_lru(key)
        return value

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._put_to_back(key, value)
        self._reorder_lru(key)
        self._enforce_lru_size()

    def __delitem__(self, key):
        super().__delitem__(key)
        self._reorder_lru(key, remove=True)

    def clear(self):
        """

        """
        super().clear()
        self._lru_list = []

    def pop(self, *args):
        """

        :param args:
        :return:
        """
        k = args[0]
        try:
            self._lru_list.remove(k)
        except ValueError:
            pass
        return super().pop(*args)

    def popitem(self):
        """

        :return:
        """
        kv_return = super().popitem()
        try:
            self._lru_list.remove(kv_return[0])
        except ValueError:
            pass
        return kv_return

    def _reorder_lru(self, key, remove=False):
        """Move key in LRU (if it exists) to 'recent' end.

        If 'delete' is True just delete from the LRU.
        """

        try:
            self._lru_list.remove(key)
        except ValueError:
            pass
        if remove:
            return
        self._lru_list.insert(0, key)

    def _enforce_lru_size(self):
        """Enforce LRU size limit in cache dictionary."""

        # if a limit was defined and we have blown it
        if self._max_lru and len(self) > self._max_lru:
            # make sure in-memory dictionary doesn't get bigger
            for key in self._lru_list[self._max_lru:]:
                super().__delitem__(key)
            # also truncate the LRU list
            self._lru_list = self._lru_list[:self._max_lru]

    #####
    # override the following two methods to implement the backing cache
    #####

    def _put_to_back(self, key, value):
        """Store 'value' in backing store, using 'key' to access."""

        pass

    def _get_from_back(self, key):
        """Retrieve value for 'key' from backing storage.

        Raises KeyError if key not in backing storage.
        """

        raise KeyError



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0
import numpy as np
from PySide6 import QtCore
from matplotlib import pyplot as plt

from VeraGrid.Gui.pandas_model import PandasModel
from VeraGrid.Gui.Main.SubClasses.Server.server import ServerMain
import VeraGrid.Gui.gui_functions as gf

from VeraGridEngine.enumerations import EngineType, BranchImpedanceMode
from VeraGridEngine.Compilers.circuit_to_data import compile_numerical_circuit_at


class CompiledArraysMain(ServerMain):
    """
    Diagrams Main
    """

    def __init__(self, parent=None):
        """

        @param parent:
        """

        # create main window
        ServerMain.__init__(self, parent=parent)

        # array modes
        self.ui.arrayModeComboBox.addItem('real')
        self.ui.arrayModeComboBox.addItem('imag')
        self.ui.arrayModeComboBox.addItem('abs')
        self.ui.arrayModeComboBox.addItem('complex')

        # Buttons
        self.ui.compute_simulation_data_pushButton.clicked.connect(self.update_islands_to_display)
        self.ui.plotArraysButton.clicked.connect(self.plot_simulation_objects_data)
        self.ui.copyArraysButton.clicked.connect(self.copy_simulation_objects_data)
        self.ui.copyArraysToNumpyButton.clicked.connect(self.copy_simulation_objects_data_to_numpy)

        # tree clicks
        self.ui.simulationDataStructuresTreeView.clicked.connect(self.view_simulation_objects_data)

    def view_simulation_objects_data(self, index):
        """
        Simulation data structure clicked
        """

        tree_mdl = self.ui.simulationDataStructuresTreeView.model()
        item = tree_mdl.itemFromIndex(index)
        path = gf.get_tree_item_path(item)

        if len(path) == 2:
            group_name = path[0]
            elm_type = path[1]

            island_idx = self.ui.simulation_data_island_comboBox.currentIndex()

            if island_idx > -1 and self.circuit.valid_for_simulation():
                # elm_type = self.ui.simulationDataStructuresTreeView.selectedIndexes()[0].data(role=QtCore.Qt.ItemDataRole.DisplayRole)

                df = self.calculation_inputs_to_display[island_idx].get_structure(elm_type)

                mdl = PandasModel(df)

                self.ui.simulationDataStructureTableView.setModel(mdl)

            else:
                self.ui.simulationDataStructureTableView.setModel(None)
        else:
            self.ui.simulationDataStructureTableView.setModel(None)

    def copy_simulation_objects_data(self):
        """
        Copy the arrays of the compiled arrays view to the clipboard
        """
        mdl = self.ui.simulationDataStructureTableView.model()
        mode = self.ui.arrayModeComboBox.currentText()
        mdl.copy_to_clipboard(mode=mode)

    def copy_simulation_objects_data_to_numpy(self):
        """
        Copy the arrays of the compiled arrays view to the clipboard
        """
        mdl = self.ui.simulationDataStructureTableView.model()
        mode = 'numpy'
        mdl.copy_to_clipboard(mode=mode)

    def plot_simulation_objects_data(self):
        """
        Plot the arrays of the compiled arrays view
        """
        mdl = self.ui.simulationDataStructureTableView.model()
        data = mdl.data_c

        # declare figure
        fig = plt.figure()
        ax1 = fig.add_subplot(111)

        if mdl.is_2d():
            ax1.spy(data)

        else:
            if mdl.is_complex():
                ax1.scatter(data.real, data.imag)
                ax1.set_xlabel('Real')
                ax1.set_ylabel('Imag')
            else:
                arr = np.arange(data.shape[0])
                ax1.scatter(arr, data)
                ax1.set_xlabel('Position')
                ax1.set_ylabel('Value')

        fig.tight_layout()
        plt.show()

    def recompile_circuits_for_display(self):
        """
        Recompile the circuits available to display
        :return:
        """
        if self.circuit is not None:

            engine = self.get_preferred_engine()

            if engine == EngineType.VeraGrid:

                if self.ui.apply_impedance_tolerances_checkBox.isChecked():
                    branch_impedance_tolerance_mode = BranchImpedanceMode.Upper
                else:
                    branch_impedance_tolerance_mode = BranchImpedanceMode.Specified

                numerical_circuit = compile_numerical_circuit_at(
                    circuit=self.circuit,
                    t_idx=None,
                    branch_tolerance_mode=branch_impedance_tolerance_mode,
                    use_stored_guess=self.ui.use_voltage_guess_checkBox.isChecked(),
                    control_taps_phase=self.ui.control_tap_phase_checkBox.isChecked(),
                    control_taps_modules=self.ui.control_tap_modules_checkBox.isChecked(),
                    control_remote_voltage=self.ui.control_remote_voltage_checkBox.isChecked(),
                )
                calculation_inputs = numerical_circuit.split_into_islands()
                self.calculation_inputs_to_display = calculation_inputs

            elif engine == EngineType.Bentayga:
                import VeraGridEngine.Compilers.circuit_to_bentayga as ben
                self.calculation_inputs_to_display = ben.get_snapshots_from_bentayga(self.circuit)

            elif engine == EngineType.NewtonPA:
                import VeraGridEngine.Compilers.circuit_to_newton_pa as ne
                self.calculation_inputs_to_display = ne.get_snapshots_from_newtonpa(self.circuit)

            else:
                # fallback to VeraGrid
                numerical_circuit = compile_numerical_circuit_at(circuit=self.circuit, t_idx=None)
                calculation_inputs = numerical_circuit.split_into_islands()
                self.calculation_inputs_to_display = calculation_inputs

            return True
        else:
            self.calculation_inputs_to_display = None
            return False

    def update_islands_to_display(self):
        """
        Compile the circuit and allow the display of the calculation objects
        :return:
        """
        self.recompile_circuits_for_display()
        self.ui.simulation_data_island_comboBox.clear()
        lst = ['Island ' + str(i) for i, circuit in enumerate(self.calculation_inputs_to_display)]
        self.ui.simulation_data_island_comboBox.addItems(lst)
        if len(self.calculation_inputs_to_display) > 0:
            self.ui.simulation_data_island_comboBox.setCurrentIndex(0)



# MIT License
#
# Copyright (c) 2018 Ross Wilson
# Copyright (c) 2024, Santiago Peñate Vera
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
A base Tiles object for pySlipQt local tiles.

All tile sources should inherit from this base class.
For example, see gmt_local.py (local tiles) and osm_tiles.py
(internet tiles).
"""

import os
import math
from typing import Tuple, Union, List
from collections.abc import Callable
from PySide6.QtGui import QPixmap
from VeraGrid.Gui.Diagrams.MapWidget.Tiles.tiles_cache import TilesCache


class BaseTiles:
    """
    A base tile object to source local tiles for pySlip.
    """

    def __init__(self,
                 levels: List[int],
                 tile_width: int,
                 tile_height: int,
                 tiles_dir: str,
                 max_lru: int):
        """
        Initialise a Tiles instance.
        :param levels: a list of level numbers that are to be served
        :param tile_width: width of each tile in pixels
        :param tile_height: height of each tile in pixels
        :param tiles_dir: path to on-disk tile cache directory
        :param max_lru: maximum number of tiles cached in-memory
        """

        # save params
        self.levels = levels
        self.tile_size_x = tile_width
        self.tile_size_y = tile_height
        self.tiles_dir = tiles_dir
        self.max_lru = max_lru

        # set min and max tile levels and current level
        self._min_level = min(self.levels)
        self._max_level = max(self.levels)
        self.level = self.min_level

        self.num_tiles_x = 0
        self.num_tiles_y = 0
        self.ppd_x = 0
        self.ppd_y = 0

        self.wrap_x = False
        self.wrap_y = False

        # set up the tile cache
        self.cache = TilesCache(tiles_dir=tiles_dir, max_lru=max_lru)

        #####
        # Now finish setting up
        #####

        # tiles extent for tile data (left, right, top, bottom)
        self.extent = (-180.0, 180.0, -85.0511, 85.0511)

        # check tile cache - we expect there to already be a directory
        if not os.path.isdir(tiles_dir):
            if os.path.isfile(tiles_dir):
                msg = ("%s doesn't appear to be a tile cache directory" % tiles_dir)
                raise Exception(msg) from None

            # raise Exception(msg) from None
            os.makedirs(tiles_dir)
            print(f"The tiles directory {tiles_dir} has been created")

    @property
    def max_level(self):
        """

        :return:
        """
        return self._max_level

    @property
    def min_level(self):
        """

        :return:
        """
        return self._min_level

    def level_in_range(self, level) -> bool:
        """
        Check if the level is in range
        :param level: level value
        :return: in range?
        """
        # first, CAN we zoom to this level?
        return self.min_level <= level <= self.max_level

    @property
    def tile_width(self):
        """

        :return:
        """
        return self.tile_size_x

    @property
    def tile_height(self):
        """

        :return:
        """
        return self.tile_size_y

    @property
    def map_width(self):
        """

        :return:
        """
        return self.num_tiles_x * self.tile_width  # virtual map width

    @property
    def map_height(self):
        """

        :return:
        """
        return self.num_tiles_y * self.tile_height  # virtual map height

    @property
    def map_llon(self):
        """

        :return:
        """
        return self.extent[0]

    @property
    def map_rlon(self):
        """

        :return:
        """
        return self.extent[1]

    @property
    def map_blat(self):
        """

        :return:
        """
        return self.extent[2]

    @property
    def map_tlat(self):
        """

        :return:
        """
        return self.extent[3]

    def set_level(self, level: int):
        """
        Prepare to serve tiles from the required level.

        :param level:  the required level

        Return True if level change occurred, else False if not possible.
        """

        # first, CAN we zoom to this level?
        if self.level_in_range(level):

            # get tile info
            info = self.GetInfo(level)
            if info is None:
                return False

            # OK, save new level
            self.level = level
            self.num_tiles_x, self.num_tiles_y, self.ppd_x, self.ppd_y = info
            return True

        else:
            return False

    def GetTile(self, x: float, y: float) -> QPixmap:
        """
        Get bitmap for tile at tile coords (x, y) and current level.

        :param x: X coord of tile required (tile coordinates)
        :param y: Y coord of tile required (tile coordinates)

        Returns bitmap object for the tile image.
        Tile coordinates are measured from map top-left.
        """

        # if we are wrapping X or Y, get wrapped tile coords
        if self.wrap_x:
            x = (x + self.num_tiles_x * self.tile_size_x) % self.num_tiles_x

        if self.wrap_y:
            y = (y + self.num_tiles_y * self.tile_size_y) % self.num_tiles_y

        try:
            # get tile from cache
            return self.cache[(self.level, x, y)]

        except KeyError as e:
            raise KeyError("Can't find tile for key '%s'" % str((self.level, x, y))) from None

    def GetInfo(self, level: int) -> Union[Tuple[float, float, None, None], None]:
        """
        Get tile info for a particular level.
        :param level:  the level to get tile info for
        Returns (num_tiles_x, num_tiles_y, ppd_x, ppd_y) or None if 'level'
        doesn't exist.

        Note that ppd_? may be meaningless for some tiles, so its
        value will be None.
        """

        # is required level available?
        if self.level_in_range(level):

            # otherwise get the information
            self.num_tiles_x = int(math.pow(2, level))
            self.num_tiles_y = int(math.pow(2, level))

            return self.num_tiles_x, self.num_tiles_y, None, None

        else:
            return None

    def GetExtent(self):
        """
        Get geo limits of the map tiles.
        (min_lon,   max_lon,   min_lat,   max_lat)

        Returns a tuple: (min_geo_x, max_geo_x, min_geo_y, max_geo_y)
        """

        return self.extent

    def tile_on_disk(self, level: int, x: float, y: float):
        """
        Return True if tile at (level, x, y) is on-disk.
        :param level: Map level
        :param x: map x coordinate
        :param y: map y coordinate
        """
        raise Exception('You must override BaseTiles.tile_on_disk(level, x, y))')

    def setCallback(self, callback: Callable[[int, float, float, QPixmap, bool], None]):
        """
        Set the "tile available" callback function.
        Only used with internet tiles.  See "tiles_net.py".
        """
        raise Exception('You must override BaseTiles.setCallback(callback))')

    def Geo2Tile(self, longitude: float, latitude: float) -> Tuple[int, int]:
        """
        Convert geo to tile fractional coordinates for level in use.
        Note that we assume the point *is* on the map!
        :param longitude: geo longitude in degrees
        :param latitude: geo latitude in degrees
        :return x_tile, y_tile (tile fractional X, Y coordinate)
        """

        raise Exception('You must override BaseTiles.Geo2Tile(longitude, latitude)')

    def Tile2Geo(self, x_tile: float, y_tile: float) -> Tuple[float, float]:
        """
        Convert tile fractional coordinates to geo for level in use.

        :param x_tile:  tile fractional X coordinate
        :param y_tile:  tile fractional Y coordinate

        Note that we assume the point *is* on the map!
        """

        raise Exception('You must override BaseTiles.Tile2Geo(x_tile, y_tile)')



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
# SPDX-License-Identifier: MPL-2.0

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from typing import List, Union, Any
import math
from PySide6 import QtGui

from VeraGridEngine.basic_structures import LogSeverity
from VeraGridEngine.Devices.multi_circuit import MultiCircuit
from VeraGridEngine.Devices.Branches.line import Line
from VeraGridEngine.Devices.Branches.transformer import Transformer2W
from VeraGridEngine.Devices.Substation.bus import Bus
from VeraGridEngine.Devices.Injections.generator import Generator
from VeraGridEngine.Devices.Injections.battery import Battery
from VeraGridEngine.Devices.Injections.static_generator import StaticGenerator
from VeraGridEngine.Devices.Injections.load import Load
from VeraGridEngine.basic_structures import Logger
from VeraGridEngine.enumerations import DeviceType
from VeraGridEngine.Devices.types import ALL_DEV_TYPES
from VeraGridEngine.Utils.NumericalMethods.numerical_stability import (sparse_instability_svd_test,
                                                                       sparse_instability_lu_test)
from VeraGridEngine.Compilers.circuit_to_data import compile_numerical_circuit_at


class GridErrorLog:
    """
    Log of grid errors
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        self.logs = dict()

        self.header = ['Object type',
                       'Name',
                       'Index',
                       'Severity',
                       'Property',
                       'Lower',
                       'Value',
                       'Upper']

    def add(self,
            object_type,
            element_name: Union[str, Any],
            element_index: int,
            severity: LogSeverity,
            propty: str,
            message: str,
            lower: Union[str, float, int] = '',
            val: Union[str, float, int] = '',
            upper: Union[str, float, int] = ''):
        """

        :param object_type:
        :param element_name:
        :param element_index:
        :param severity:
        :param propty:
        :param message:
        :param lower:
        :param val:
        :param upper:
        :return:
        """

        e = [object_type,
             str(element_name),
             element_index,
             severity,
             propty,
             str(lower),
             str(val),
             str(upper)]

        if message in self.logs.keys():
            self.logs[message].append(e)
        else:
            self.logs[message] = [e]

    def clear(self):
        """
        Delete all logs
        """
        self.logs.clear()

    def add_normal_logger(self, logger: Logger):
        """
        Add normal logger
        :param logger: Logger
        """
        for l in logger.entries:
            self.add(object_type=l.device_class,
                     element_name=l.device,
                     element_index=0,
                     severity=l.severity,
                     propty=l.device_property,
                     message=l.msg,
                     val=l.value,
                     lower=l.expected_object_value,
                     upper=l.expected_value)

    def get_model(self) -> "QtGui.QStandardItemModel":
        """
        Get TreeView Model
        :return: QStandardItemModel
        """
        model = QtGui.QStandardItemModel()
        model.setHorizontalHeaderLabels(self.header)

        # populate data
        for message_key, entries in self.logs.items():
            parent1 = QtGui.QStandardItem(message_key)
            for object_type, element_name, element_index, severity, prop, lower, val, upper in entries:
                children = [QtGui.QStandardItem(str(object_type)),
                            QtGui.QStandardItem(str(element_name)),
                            QtGui.QStandardItem(str(element_index)),
                            QtGui.QStandardItem(severity.value),
                            QtGui.QStandardItem(str(prop)),
                            QtGui.QStandardItem(str(lower)),
                            QtGui.QStandardItem(str(val)),
                            QtGui.QStandardItem(str(upper))]
                for chld in children:
                    chld.setEditable(False)

                parent1.appendRow(children)

            parent1.setEditable(False)
            model.appendRow(parent1)

        return model

    def get_df(self):
        """
        Save analysis to excel
        :return:
        """
        data = list()

        for message in self.logs.keys():

            items = self.logs[message]

            for [object_type, element_name, element_index, severity, propty, lower, val, upper] in items:
                data.append([message, object_type, element_name, element_index, severity, propty, lower, val, upper])

        hdr = ['Message', 'Object type', 'Name', 'Index', 'Severity', 'Property', 'Lower', 'Value', 'Upper']
        return pd.DataFrame(data=data, columns=hdr)

    def save(self, filename):
        """
        Save analysis to excel
        :param filename:
        :return:
        """
        df = self.get_df()
        df.to_excel(filename)


class FixableErrorOutOfRange:
    """
    Error type for when a value is out of range
    """

    def __init__(self,
                 grid_element: ALL_DEV_TYPES,
                 property_name: str,
                 value: float,
                 lower_limit: float,
                 upper_limit: float):
        """

        :param grid_element:
        :param property_name:
        :param value:
        :param lower_limit:
        :param upper_limit:
        """
        self.grid_element: ALL_DEV_TYPES = grid_element
        self.property_name: str = property_name
        self.value: float = value
        self.lower_limit: float = lower_limit
        self.upper_limit: float = upper_limit

    def fix(self, logger: Logger = Logger(), fix_ts=False):
        """

        :param logger:
        :param fix_ts:
        :return:
        """
        if self.value < self.lower_limit:
            self.grid_element.set_snapshot_value(property_name=self.property_name, value=self.lower_limit)
            logger.add_info("Fixed " + self.property_name, device=self.grid_element.idtag, value=self.value)

        elif self.value > self.upper_limit:
            self.grid_element.set_snapshot_value(property_name=self.property_name, value=self.upper_limit)
            logger.add_info("Fixed " + self.property_name, device=self.grid_element.idtag, value=self.value)

        # fix the associated time series
        gc_prop = self.grid_element.registered_properties.get(self.property_name, None)
        if fix_ts and gc_prop.has_profile():
            profile = self.grid_element.get_profile_by_prop(prop=gc_prop)
            for i in range(profile.size()):
                value = profile[i]
                if value < self.lower_limit:
                    profile[i] = self.lower_limit
                    logger.add_info("Fixed " + self.property_name, device=self.grid_element.idtag, value=value)

                elif value > self.upper_limit:
                    profile[i] = self.upper_limit
                    logger.add_info("Fixed " + self.property_name, device=self.grid_element.idtag, value=value)


class FixableErrorRangeFlip:
    """
    Error type for when a range is reversed
    """

    def __init__(self, grid_element, property_name_low, property_name_high, value_low, value_high):
        """

        :param grid_element:
        :param property_name_low:
        :param property_name_high:
        :param value_low:
        :param value_high:
        """
        self.grid_element = grid_element
        self.property_name_low = property_name_low
        self.property_name_high = property_name_high
        self.value_low = value_low
        self.value_high = value_high

    def fix(self, logger: Logger = Logger(), fix_ts=False):
        """

        :param logger:
        :param fix_ts:
        """
        if self.value_high < self.value_low:
            # flip the values
            self.grid_element.set_snapshot_value(property_name=self.property_name_low, value=self.value_high)
            self.grid_element.set_snapshot_value(property_name=self.property_name_high, value=self.value_low)


class FixableErrorValueCorrection:
    """
    Error type for when a range is reversed
    """

    def __init__(self, grid_element, property_name_low, property_name_high, value_low, value_high):
        """

        :param grid_element:
        :param property_name_low:
        :param property_name_high:
        :param value_low:
        :param value_high:
        """
        self.grid_element = grid_element
        self.property_name_low = property_name_low
        self.property_name_high = property_name_high
        self.value_low = value_low
        self.value_high = value_high

    def fix(self, logger: Logger = Logger(), fix_ts=False):
        """

        :param logger:
        :param fix_ts:
        """

        self.grid_element.set_snapshot_value(property_name=self.property_name_low, value=self.value_low)
        self.grid_element.set_snapshot_value(property_name=self.property_name_high, value=self.value_high)


class FixableErrorNegative:
    """
    Error type for when a value is negative
    """

    def __init__(self, grid_element, property_name, value):
        """

        :param grid_element:
        :param property_name:
        :param value:
        """
        self.grid_element = grid_element
        self.property_name = property_name
        self.value = value

    def fix(self, logger: Logger = Logger(), fix_ts=False):
        """

        :param logger:
        :param fix_ts:
        :return:
        """
        # set the same value but positive
        if self.value < 0:
            self.grid_element.set_snapshot_value(property_name=self.property_name, value=-self.value)


class FixableTransformerVtaps:
    """
    Error type for when a transformer virtual taps are wrong
    """

    def __init__(self, grid_element, maximum_difference):
        """

        :param grid_element:
        :param maximum_difference:
        """
        self.grid_element = grid_element
        self.maximum_difference = maximum_difference

    def fix(self, logger: Logger = Logger(), fix_ts=False):
        """

        :param logger:
        :param fix_ts:
        :return:
        """
        # set the same value but positive
        self.grid_element.fix_inconsistencies(logger=logger,
                                              maximum_difference=self.maximum_difference)


FIXABLE_ERROR_TYPES = Union[
    FixableTransformerVtaps,
    FixableErrorOutOfRange,
    FixableErrorRangeFlip,
    FixableErrorNegative,
    FixableErrorValueCorrection
]


def analyze_lines(elements: List[Line],
                  object_type: DeviceType,
                  branch_connection_voltage_tolerance: float,
                  eps_min: float,
                  eps_max: float,
                  branch_x_threshold: float,
                  logger: GridErrorLog,
                  fixable_errors: List[FIXABLE_ERROR_TYPES]):
    """

    :param elements:
    :param object_type:
    :param branch_connection_voltage_tolerance:
    :param eps_min:
    :param eps_max:
    :param branch_x_threshold:
    :param logger:
    :param fixable_errors:
    :return:
    """
    for i, elm in enumerate(elements):

        V1 = min(elm.bus_to.Vnom, elm.bus_from.Vnom)
        V2 = max(elm.bus_to.Vnom, elm.bus_from.Vnom)

        if elm.bus_from is None:
            logger.add(object_type=elm.device_type.value,
                       element_name=elm,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty="bus_from",
                       message="The bus from is None")

        if elm.bus_to is None:
            logger.add(object_type=elm.device_type.value,
                       element_name=elm,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty="bus_to",
                       message="The bus to is None")

        if elm.bus_to == elm.bus_from:
            logger.add(object_type=elm.device_type.value,
                       element_name=elm,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty="bus_from, bus_to",
                       message="Branch connected in loop (bus_from = bus_to)")

        if V1 > 0 and V2 > 0:
            per = V1 / V2
            if per < (1.0 - branch_connection_voltage_tolerance):
                logger.add(object_type=object_type.value,
                           element_name=elm.name,
                           element_index=i,
                           severity=LogSeverity.Error,
                           propty='Connection',
                           message='The branch is connected between voltages '
                                   'that differ in {}% or more. Should this '
                                   'be a transformer?'.format(int(branch_connection_voltage_tolerance * 100)),
                           lower=V1,
                           upper=V2)
        else:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='Voltage',
                       message='The branch does is connected to a bus with Vnom=0, this is terrible.')

        if elm.name == '':
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Information,
                       propty='name',
                       message='The branch does not have a name')

        if elm.rate < 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='rate',
                       message='The rating is negative. This cannot be.',
                       lower="0",
                       val=elm.rate)
            fixable_errors.append(FixableErrorNegative(grid_element=elm,
                                                       property_name='rate',
                                                       value=elm.rate))

        elif elm.rate == 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='rate',
                       message='There is no nominal power, this is bad.',
                       val=elm.rate)

        if elm.R < 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='R',
                       message='The resistance is negative, that cannot be.',
                       lower="0.0",
                       val=elm.R)
            fixable_errors.append(FixableErrorNegative(grid_element=elm,
                                                       property_name='R',
                                                       value=elm.R))

        elif elm.R == 1e-20:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Information,
                       propty='R',
                       message='The resistance is practically zero.',
                       val=elm.R)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='R',
                                                         value=elm.R,
                                                         lower_limit=eps_min,
                                                         upper_limit=eps_max))

        if elm.X == 1e-20:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='X',
                       message='The reactance is practically zero. This hurts numerical conditioning.',
                       val=elm.X)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='X',
                                                         value=elm.X,
                                                         lower_limit=eps_min,
                                                         upper_limit=eps_max))

        if elm.B == 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='B',
                       message='There is no susceptance, this could hurt numerical conditioning.',
                       val=elm.B)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='B',
                                                         value=elm.B,
                                                         lower_limit=eps_min,
                                                         upper_limit=eps_max))

        # check the reactance
        if elm.X < branch_x_threshold:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Warning,
                       propty='X',
                       message='Line X is too low and can make models numerically unstable',
                       lower=str(branch_x_threshold),
                       upper="1e20",
                       val=elm.X)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='X',
                                                         value=elm.X,
                                                         lower_limit=branch_x_threshold,
                                                         upper_limit=1e20))


def analyze_transformers(elements: List[Transformer2W],
                         object_type: DeviceType,
                         transformer_virtual_tap_tolerance: float,
                         eps_min: float,
                         eps_max: float,
                         min_vcc: float,
                         max_vcc: float,
                         tap_min: float,
                         tap_max: float,
                         branch_x_threshold: float,
                         logger: GridErrorLog,
                         fixable_errors: List[FIXABLE_ERROR_TYPES]):
    """

    :param elements:
    :param object_type:
    :param transformer_virtual_tap_tolerance:
    :param eps_min:
    :param eps_max:
    :param min_vcc:
    :param max_vcc:
    :param tap_min:
    :param tap_max:
    :param branch_x_threshold:
    :param logger:
    :param fixable_errors:
    :return:
    """
    for i, elm in enumerate(elements):

        if elm.bus_from is None:
            logger.add(object_type=elm.device_type.value,
                       element_name=elm,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty="bus_from",
                       message="The bus from is None")

        if elm.bus_to is None:
            logger.add(object_type=elm.device_type.value,
                       element_name=elm,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty="bus_to",
                       message="The bus to is None")

        if elm.bus_to == elm.bus_from:
            logger.add(object_type=elm.device_type.value,
                       element_name=elm,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty="bus_from, bus_to",
                       message="Branch connected in loop (bus_from = bus_to)")

        if elm.name == '':
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='name',
                       message='The branch does not have a name')

        if elm.rate <= 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='rate',
                       message='The rating is negative. This cannot be.',
                       lower="0",
                       val=elm.rate)
            fixable_errors.append(FixableErrorNegative(grid_element=elm,
                                                       property_name='rate',
                                                       value=elm.rate))

        # check R and X
        if elm.R == 0.0 and elm.X == 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='R+X',
                       message='There is no impedance, set at least a very low value',
                       lower="0",
                       val=elm.R + elm.X)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='R',
                                                         value=elm.rate,
                                                         lower_limit=eps_min,
                                                         upper_limit=eps_max))
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='X',
                                                         value=elm.rate,
                                                         lower_limit=eps_min,
                                                         upper_limit=eps_max))

        else:
            if elm.R < 0.0:
                logger.add(object_type=object_type.value,
                           element_name=elm.name,
                           element_index=i,
                           severity=LogSeverity.Warning,
                           propty='R',
                           message='The resistance is negative, that cannot be.',
                           lower="0",
                           val=elm.R)
                fixable_errors.append(FixableErrorNegative(grid_element=elm,
                                                           property_name='R',
                                                           value=elm.R))

            elif elm.R == 0.0:
                logger.add(object_type=object_type.value,
                           element_name=elm.name,
                           element_index=i,
                           severity=LogSeverity.Information,
                           propty='R',
                           message='The resistance is exactly zero',
                           val=elm.R)
                fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                             property_name='R',
                                                             value=elm.R,
                                                             lower_limit=eps_min,
                                                             upper_limit=eps_max))

            elif elm.X == 0.0:
                logger.add(object_type=object_type.value,
                           element_name=elm.name,
                           element_index=i,
                           severity=LogSeverity.Information,
                           propty='X',
                           message='The reactance is exactly zero',
                           val=elm.X)
                fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                             property_name='X',
                                                             value=elm.rate,
                                                             lower_limit=eps_min,
                                                             upper_limit=eps_max))

        # check tap module
        if elm.tap_module > tap_max:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Warning,
                       propty='tap_module',
                       message='Tap module too high',
                       upper=str(tap_max),
                       val=elm.tap_module)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='tap_module',
                                                         value=elm.tap_module,
                                                         lower_limit=tap_min,
                                                         upper_limit=tap_max))

        elif elm.tap_module < tap_min:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Warning,
                       propty='tap_module',
                       message='Tap module too low',
                       upper=str(tap_min),
                       val=elm.tap_module)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='tap_module',
                                                         value=elm.tap_module,
                                                         lower_limit=tap_min,
                                                         upper_limit=tap_max))

        # check virtual taps
        tap_f, tap_t = elm.get_virtual_taps()

        if (1.0 - transformer_virtual_tap_tolerance) > tap_f or tap_f > (
                1.0 + transformer_virtual_tap_tolerance):
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='HV or LV',
                       message='Large nominal voltage mismatch at the "from" bus',
                       lower=str(1.0 - transformer_virtual_tap_tolerance),
                       val=str(tap_f),
                       upper=str(1.0 + transformer_virtual_tap_tolerance))
            fixable_errors.append(FixableTransformerVtaps(grid_element=elm,
                                                          maximum_difference=transformer_virtual_tap_tolerance))

        if (1.0 - transformer_virtual_tap_tolerance) > tap_t or tap_t > (
                1.0 + transformer_virtual_tap_tolerance):
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='HV or LV',
                       message='Large nominal voltage mismatch at the "to" bus',
                       lower=str(1.0 - transformer_virtual_tap_tolerance),
                       val=str(tap_t),
                       upper=str(1.0 + transformer_virtual_tap_tolerance))
            fixable_errors.append(FixableTransformerVtaps(grid_element=elm,
                                                          maximum_difference=transformer_virtual_tap_tolerance))

        # check VCC
        vcc = elm.get_vcc()
        if vcc < min_vcc or vcc > max_vcc:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Warning,
                       propty='Vcc',
                       message='The short circuit value is suspicious',
                       lower=str(min_vcc),
                       upper=str(max_vcc),
                       val=str(vcc))

        # check the nominal power
        if elm.Sn > 0:
            sn_ratio = elm.rate / elm.Sn
            if not (0.9 < sn_ratio < 1.1):
                logger.add(object_type=object_type.value,
                           element_name=elm.name,
                           element_index=i,
                           severity=LogSeverity.Warning,
                           propty='rate / Sn',
                           message='Transformer rating is too different from the nominal power',
                           lower=str(elm.Sn * 0.9),
                           upper=str(elm.Sn * 1.1),
                           val=elm.rate)
                fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                             property_name='rate',
                                                             value=elm.rate,
                                                             lower_limit=elm.Sn,
                                                             upper_limit=elm.Sn))

        # check the reactance
        if elm.X < branch_x_threshold:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Warning,
                       propty='X',
                       message='Transformer X is too low and can make models numerically unstable',
                       lower=str(branch_x_threshold),
                       upper="1e20",
                       val=elm.X)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='X',
                                                         value=elm.X,
                                                         lower_limit=branch_x_threshold,
                                                         upper_limit=1e20))


def analyze_buses(elements: List[Bus],
                  object_type: DeviceType,
                  tap_min: float,
                  tap_max: float,
                  v_low: float,
                  v_high: float,
                  logger: GridErrorLog,
                  fixable_errors: List[FIXABLE_ERROR_TYPES]):
    """

    :param elements:
    :param object_type:
    :param tap_min:
    :param tap_max:
    :param v_low:
    :param v_high:
    :param logger:
    :param fixable_errors:
    :return:
    """
    names = set()

    for i, elm in enumerate(elements):

        if elm.Vnom <= 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='Vnom',
                       message='The nominal voltage is <= 0, this causes problems',
                       lower="0",
                       val=elm.Vnom)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='Vnom',
                                                         value=elm.Vnom,
                                                         lower_limit=tap_min,
                                                         upper_limit=tap_max))

        if elm.Vmin == 0.0:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='Vmin',
                       message='Vmin = 0, that affects the nonlinear OPF',
                       lower="0",
                       val=elm.Vmin)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='Vmin',
                                                         value=elm.Vmin,
                                                         lower_limit=v_low,
                                                         upper_limit=v_low))

        if elm.Vmax < elm.Vmin or elm.Vmax == 0:

            if elm.Vmax == 0:
                logger.add(object_type=object_type.value,
                           element_name=elm.name,
                           element_index=i,
                           severity=LogSeverity.Error,
                           propty='Vmax',
                           message=' Vmax = 0, that affects the nonlinear OPF',
                           lower="0",
                           val=elm.Vmax)

            if elm.Vmax < elm.Vmin:
                logger.add(object_type=object_type.value,
                           element_name=elm.name,
                           element_index=i,
                           severity=LogSeverity.Error,
                           propty='Vmax',
                           message=' Vmax <= Vmin, that affects the nonlinear OPF',
                           lower="0",
                           val=elm.Vmax)

            fixable_errors.append(FixableErrorOutOfRange(grid_element=elm,
                                                         property_name='Vmax',
                                                         value=elm.Vmax,
                                                         lower_limit=v_high,
                                                         upper_limit=v_high))

        if elm.name == '':
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Error,
                       propty='name',
                       message='The bus does not have a name')

        if elm.name in names:
            logger.add(object_type=object_type.value,
                       element_name=elm.name,
                       element_index=i,
                       severity=LogSeverity.Information,
                       propty='name',
                       message='The bus name is not unique')

        # add the name to a set
        names.add(elm.name)


def analyze_generators(elements: List[Generator],
                       object_type: DeviceType,
                       time_profile,
                       v_low: float,
                       v_high: float,
                       logger: GridErrorLog,
                       fixable_errors: List[FIXABLE_ERROR_TYPES]):
    """

    :param elements:
    :param object_type:
    :param time_profile:
    :param v_low:
    :param v_high:
    :param logger:
    :param fixable_errors:
    :return:
    """
    Pg = 0
    Pg_prof = 0

    for k, obj in enumerate(elements):
        Pg += obj.P * obj.active

        if time_profile is not None:
            Pg_prof += obj.P_prof.toarray() * obj.active_prof.toarray()

        if obj.bus is None:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty="bus",
                       message="The bus is None")

        if obj.Vset < v_low:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Warning,
                       propty='Vset',
                       message='The set point looks too low',
                       lower=str(v_low),
                       val=obj.Vset)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=obj,
                                                         property_name='Vset',
                                                         value=obj.Vset,
                                                         lower_limit=v_low,
                                                         upper_limit=v_high))

        elif obj.Vset > v_high:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Warning,
                       propty='Vset',
                       message='The set point looks too high',
                       val=obj.Vset,
                       upper=str(v_high))
            fixable_errors.append(FixableErrorOutOfRange(grid_element=obj,
                                                         property_name='Vset',
                                                         value=obj.Vset,
                                                         lower_limit=v_low,
                                                         upper_limit=v_high))

        if obj.Qmax < obj.Qmin:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty='Qmax < Qmin',
                       message='Wrong Q limit bounds',
                       upper=obj.Qmax,
                       lower=obj.Qmin)
            fixable_errors.append(FixableErrorRangeFlip(grid_element=obj,
                                                        property_name_low='Qmin',
                                                        property_name_high='Qmax',
                                                        value_low=obj.Qmin,
                                                        value_high=obj.Qmax))

        # elif obj.Qmax == obj.Qmin and obj.is_controlled:
        elif obj.Qmax == obj.Qmin:

            if obj.Qmax == 0.0:
                propty = 'Qmax == Qmin == 0'
                message = 'Q limits equal to zero'

            else:
                propty = 'Qmax == Qmin'
                message = 'Inflexible Q limit bounds'

            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty=propty,
                       message=message,
                       upper=obj.Qmax,
                       lower=obj.Qmin)
            val = np.round(obj.Snom * 0.8, 2)
            fixable_errors.append(FixableErrorValueCorrection(grid_element=obj,
                                                              property_name_low='Qmin',
                                                              property_name_high='Qmax',
                                                              value_low=-val,
                                                              value_high=val))

        if obj.Pmax < obj.Pmin:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty='Pmax < Pmin',
                       message='Wrong P limit bounds',
                       upper=obj.Pmax,
                       lower=obj.Pmin)
            fixable_errors.append(FixableErrorRangeFlip(grid_element=obj,
                                                        property_name_low='Pmin',
                                                        property_name_high='Pmax',
                                                        value_low=obj.Pmin,
                                                        value_high=obj.Pmax))
        elif obj.Pmax == obj.Pmin:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty='Pmax == Pmin',
                       message='Inflexible P limit bounds',
                       upper=obj.Pmax,
                       lower=obj.Pmin)
            fixable_errors.append(FixableErrorValueCorrection(grid_element=obj,
                                                              property_name_low='Pmin',
                                                              property_name_high='Pmax',
                                                              value_low=0,
                                                              value_high=obj.Snom))

    return Pg, Pg_prof


def analyze_batteries(elements: List[Battery],
                      object_type: DeviceType,
                      time_profile,
                      v_low: float,
                      v_high: float,
                      logger: GridErrorLog,
                      fixable_errors: List[FIXABLE_ERROR_TYPES]):
    """

    :param elements:
    :param object_type:
    :param time_profile:
    :param v_low:
    :param v_high:
    :param logger:
    :param fixable_errors:
    :return:
    """
    Pg = 0
    Pg_prof = 0

    for k, obj in enumerate(elements):
        Pg += obj.P * obj.active

        if obj.bus is None:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty="bus",
                       message="The bus is None")

        if time_profile is not None:
            Pg_prof += obj.P_prof.toarray() * obj.active_prof.toarray()

        if obj.Vset < v_low:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Warning,
                       propty='Vset',
                       message='The set point looks too low',
                       lower=str(v_low),
                       val=obj.Vset)
            fixable_errors.append(FixableErrorOutOfRange(grid_element=obj,
                                                         property_name='Vset',
                                                         value=obj.Vset,
                                                         lower_limit=v_low,
                                                         upper_limit=v_high))

        elif obj.Vset > v_high:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Warning,
                       propty='Vset',
                       message='The set point looks too high',
                       val=obj.Vset,
                       upper=str(v_high))
            fixable_errors.append(FixableErrorOutOfRange(grid_element=obj,
                                                         property_name='Vset',
                                                         value=obj.Vset,
                                                         lower_limit=v_low,
                                                         upper_limit=v_high))

        if obj.Qmax < obj.Qmin:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty='Qmax < Qmin',
                       message='Wrong Q limit bounds',
                       upper=obj.Qmax,
                       lower=obj.Qmin)
            fixable_errors.append(FixableErrorRangeFlip(grid_element=obj,
                                                        property_name_low='Qmin',
                                                        property_name_high='Qmax',
                                                        value_low=obj.Qmin,
                                                        value_high=obj.Qmax))

        if obj.Pmax < obj.Pmin:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty='Pmax < Pmin',
                       message='Wrong P limit bounds',
                       upper=obj.Pmax,
                       lower=obj.Pmin)
            fixable_errors.append(FixableErrorRangeFlip(grid_element=obj,
                                                        property_name_low='Pmin',
                                                        property_name_high='Pmax',
                                                        value_low=obj.Pmin,
                                                        value_high=obj.Pmax))

        if obj.max_soc < obj.min_soc:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty='max_soc < min_soc',
                       message='Wrong SoC limit bounds',
                       upper=obj.max_soc,
                       lower=obj.min_soc)
            fixable_errors.append(FixableErrorRangeFlip(grid_element=obj,
                                                        property_name_low='min_soc',
                                                        property_name_high='max_soc',
                                                        value_low=obj.min_soc,
                                                        value_high=obj.max_soc))

        if obj.Enom <= 0:
            logger.add(object_type=object_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty='Enom',
                       message='Invalid nominal energy',
                       lower="0",
                       val=obj.Enom)

    return Pg, Pg_prof


def analyze_static_gen(elements: List[StaticGenerator],
                       time_profile,
                       logger: GridErrorLog, ):
    """

    :param elements:
    :param time_profile:
    :param logger:
    :return:
    """
    Pg = 0.0
    Qg = 0.0
    Pg_prof = 0.0
    Qg_prof = 0.0

    for k, obj in enumerate(elements):
        Pg += obj.P * obj.active
        Qg += obj.Q * obj.active

        if obj.bus is None:
            logger.add(object_type=obj.device_type.value,
                       element_name=obj,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty="bus",
                       message="The bus is None")

        if time_profile is not None:
            Pg_prof += obj.P_prof.toarray() * obj.active_prof.toarray()
            Qg_prof += obj.Q_prof.toarray() * obj.active_prof.toarray()

    return Pg, Qg, Pg_prof, Qg_prof


def analyze_load(elements: List[Load],
                 time_profile,
                 logger: GridErrorLog, ):
    """

    :param elements:
    :param time_profile:
    :param logger:
    :return:
    """
    Pl = 0.0
    Ql = 0.0
    Pl_prof = 0.0
    Ql_prof = 0.0

    for k, elm in enumerate(elements):
        Pl += elm.P * elm.active
        Ql += elm.Q * elm.active

        if elm.bus is None:
            logger.add(object_type=elm.device_type.value,
                       element_name=elm,
                       element_index=k,
                       severity=LogSeverity.Error,
                       propty="bus",
                       message="The bus is None")

        if time_profile is not None:
            Pl_prof += elm.P_prof.toarray() * elm.active_prof.toarray()
            Ql_prof += elm.Q_prof.toarray() * elm.active_prof.toarray()

    return Pl, Ql, Pl_prof, Ql_prof


def grid_analysis(circuit: MultiCircuit,
                  analyze_ts=True,
                  imbalance_threshold=0.02,
                  v_low=0.95,
                  v_high=1.05,
                  tap_min=0.95,
                  tap_max=1.05,
                  transformer_virtual_tap_tolerance=0.1,
                  branch_connection_voltage_tolerance=0.1,
                  min_vcc=8,
                  max_vcc=18,
                  logger=GridErrorLog(),
                  branch_x_threshold=1e-4,
                  condition_number_threshold=1e4,
                  eps_max: float = 1e20,
                  eps_min: float = 1e-20):
    """
    Analyze the model data
    :param circuit: Circuit to analyze
    :param analyze_ts: Analyze time series
    :param imbalance_threshold: Allowed percentage of imbalance
    :param v_low: lower voltage setting
    :param v_high: higher voltage setting
    :param tap_min: minimum tap value
    :param tap_max: maximum tap value
    :param transformer_virtual_tap_tolerance:
    :param branch_connection_voltage_tolerance:
    :param max_vcc: maximum short circuit voltage (%)
    :param min_vcc: Minimum short circuit voltage (%)
    :param logger: GridErrorLog
    :param branch_x_threshold: Value to compare branches X such that it is numerically stable
    :param condition_number_threshold: Condition number threshold to report unstability
    :param eps_max: Max epsylon value for comparison
    :param eps_min: Min epsylon value for comparison
    :return: list of fixable error objects
    """

    # check for duplicated uuid's
    duplicates_logger = Logger()
    _, ok_dict = circuit.get_all_elements_dict(logger=duplicates_logger)
    logger.add_normal_logger(duplicates_logger)

    if circuit.time_profile is not None:
        nt = len(circuit.time_profile)
    else:
        nt = 0

    fixable_errors: List[FIXABLE_ERROR_TYPES] = list()

    Pl = 0
    Pg = 0
    Pl_prof = np.zeros(nt)
    Pg_prof = np.zeros(nt)

    Ql = 0
    Qg = 0
    Ql_prof = np.zeros(nt)
    Qg_prof = np.zeros(nt)

    Qmin = 0.0
    Qmax = 0.0

    analyze_buses(elements=circuit.get_buses(),
                  object_type=DeviceType.BusDevice,
                  tap_min=tap_min,
                  tap_max=tap_max,
                  logger=logger,
                  v_low=v_low,
                  v_high=v_high,
                  fixable_errors=fixable_errors)

    analyze_lines(elements=circuit.get_lines(),
                  object_type=DeviceType.LineDevice,
                  branch_connection_voltage_tolerance=branch_connection_voltage_tolerance,
                  eps_min=eps_min,
                  eps_max=eps_max,
                  branch_x_threshold=branch_x_threshold,
                  logger=logger,
                  fixable_errors=fixable_errors)

    analyze_transformers(elements=circuit.get_transformers2w(),
                         object_type=DeviceType.Transformer2WDevice,
                         transformer_virtual_tap_tolerance=transformer_virtual_tap_tolerance,
                         eps_min=eps_min,
                         eps_max=eps_max,
                         min_vcc=min_vcc,
                         max_vcc=max_vcc,
                         tap_min=tap_min,
                         tap_max=tap_max,
                         branch_x_threshold=branch_x_threshold,
                         logger=logger,
                         fixable_errors=fixable_errors)

    analyze_transformers(elements=circuit.get_windings(),
                         object_type=DeviceType.WindingDevice,
                         transformer_virtual_tap_tolerance=transformer_virtual_tap_tolerance,
                         eps_min=eps_min,
                         eps_max=eps_max,
                         min_vcc=min_vcc,
                         max_vcc=max_vcc,
                         tap_min=tap_min,
                         tap_max=tap_max,
                         branch_x_threshold=branch_x_threshold,
                         logger=logger,
                         fixable_errors=fixable_errors)

    Pgg, Pgg_prof = analyze_generators(elements=circuit.get_generators(),
                                       object_type=DeviceType.GeneratorDevice,
                                       time_profile=circuit.get_time_array(),
                                       v_low=v_low,
                                       v_high=v_high,
                                       logger=logger,
                                       fixable_errors=fixable_errors)
    Pg += Pgg
    Pg_prof += Pgg_prof

    Pgb, Pgb_prof = analyze_batteries(elements=circuit.get_batteries(),
                                      object_type=DeviceType.GeneratorDevice,
                                      time_profile=circuit.get_time_array(),
                                      v_low=v_low,
                                      v_high=v_high,
                                      logger=logger,
                                      fixable_errors=fixable_errors)

    Pg += Pgb
    Pg_prof += Pgb_prof

    Pgs, Qgs, Pgs_prof, Qgs_prof = analyze_static_gen(elements=circuit.get_static_generators(),
                                                      time_profile=circuit.get_time_array(),
                                                      logger=logger)

    Pg += Pgs
    Pg_prof += Pgs_prof
    Qg += Qgs
    Qg_prof += Qgs_prof

    Pll, Qll, Pll_prof, Qll_prof = analyze_load(elements=circuit.get_loads(),
                                                time_profile=circuit.get_time_array(),
                                                logger=logger)

    Pl += Pll
    Pl_prof += Pll_prof
    Ql += Qll
    Ql_prof += Qll_prof

    # compare loads
    p_ratio = abs(Pl - Pg) / (Pl + eps_min)

    if p_ratio > imbalance_threshold:
        msg = ">> " + str(imbalance_threshold) + "%"
        logger.add(object_type='Grid snapshot',
                   element_name=circuit,
                   element_index=-1,
                   severity=LogSeverity.Error,
                   propty='Active power balance ' + msg,
                   message='There is too much active power imbalance',
                   val="{:.1f}".format(p_ratio * 100))

    # compare reactive power limits
    if not (Qmin <= -Ql <= Qmax):
        logger.add(object_type='Grid snapshot',
                   element_name=circuit,
                   element_index=-1,
                   severity=LogSeverity.Error,
                   propty="Reactive power out of bounds",
                   message='There is too much reactive power imbalance',
                   lower=str(Qmin),
                   val=str(Ql),
                   upper=str(Qmax))

    # analyze the time series data
    if analyze_ts:
        if circuit.time_profile is not None:
            nt = len(circuit.time_profile)

        for t in range(nt):
            # compare loads
            p_ratio = abs(Pl_prof[t] - Pg_prof[t]) / (Pl_prof[t] + eps_min)
            if p_ratio > imbalance_threshold:
                msg = ">> " + str(imbalance_threshold) + "%"
                logger.add(object_type='Active power balance',
                           element_name=circuit,
                           element_index=t,
                           severity=LogSeverity.Error,
                           propty=msg,
                           message='There is too much active power imbalance',
                           val="{:.1f}".format(p_ratio * 100))

            # compare reactive power limits
            if not (Qmin <= -Ql_prof[t] <= Qmax):
                logger.add(object_type='Reactive power power balance',
                           element_name=circuit,
                           element_index=t,
                           severity=LogSeverity.Error,
                           propty="Reactive power out of bounds",
                           message='There is too much reactive power imbalance',
                           lower=str(Qmin),
                           val=float(Ql_prof[t]),
                           upper=str(Qmax))

    # analyze the numerical stability
    nc = compile_numerical_circuit_at(circuit, t_idx=None)  # compile the snapshot

    Sbus = nc.get_power_injections()
    indices = nc.get_simulation_indices(Sbus=Sbus)
    lin_adm = nc.get_linear_admittance_matrices(indices=indices)
    rcond, unstable = sparse_instability_svd_test(lin_adm.Bbus, condition_number_thrshold=1.0 / condition_number_threshold)

    if unstable:
        logger.add(object_type='matrix',
                   element_name=circuit,
                   element_index=-1,
                   severity=LogSeverity.Error,
                   propty="condition number",
                   message='B matrix is SVD-Unstable: this may make linear power flows output nonsense',
                   lower="",
                   val=str(rcond),
                   upper=str(1.0 / condition_number_threshold))

    rcond, unstable = sparse_instability_lu_test(lin_adm.get_Bred(pqpv=indices.no_slack),
                                                 condition_number_thrshold=condition_number_threshold)

    if unstable:
        logger.add(object_type='matrix',
                   element_name=circuit,
                   element_index=-1,
                   severity=LogSeverity.Error,
                   propty="condition number",
                   message='B matrix is LU-Unstable: this may make linear power flows output nonsense',
                   lower="",
                   val=str(rcond),
                   upper=str(condition_number_threshold))

    return fixable_errors


def object_histogram_analysis(circuit: MultiCircuit,
                              object_type: DeviceType,
                              t_idx: Union[None, int],
                              fig=None):
    """
    Draw the histogram analysis of the provided object type
    :param circuit: Circuit
    :param object_type: Object Type (DeviceType)
    :param t_idx: Time index (None or int) to get the data
    :param fig: matplotlib figure (if None, a new one is created)
    """

    if object_type == DeviceType.LineDevice.value:
        properties = ['R', 'X', 'B', 'rate']
        log_scale = [False, False, False, False, False]
        objects = circuit.lines

    elif object_type == DeviceType.Transformer2WDevice.value:
        properties = ['R', 'X', 'G', 'B', 'tap_module', 'tap_phase', 'rate']
        log_scale = [False, False, False, False, False, False, False]
        objects = circuit.transformers2w

    elif object_type == DeviceType.BusDevice.value:
        properties = ['Vnom']
        log_scale = [False]
        objects = circuit.get_buses()

    elif object_type == DeviceType.GeneratorDevice.value:
        properties = ['Vset', 'P', 'Qmin', 'Qmax']
        log_scale = [False, False, False, False]
        objects = circuit.get_generators()

    elif object_type == DeviceType.BatteryDevice.value:
        properties = ['Vset', 'P', 'Qmin', 'Qmax']
        log_scale = [False, False, False, False]
        objects = circuit.get_batteries()

    elif object_type == DeviceType.StaticGeneratorDevice.value:
        properties = ['P', 'Q']
        log_scale = [False, False]
        objects = circuit.get_static_generators()

    elif object_type == DeviceType.ShuntDevice.value:
        properties = ['G', 'B']
        log_scale = [False, False]
        objects = circuit.get_shunts()

    elif object_type == DeviceType.LoadDevice.value:
        properties = ['P', 'Q', 'Ir', 'Ii', 'G', 'B']
        log_scale = [False, False, False, False, False, False]
        objects = circuit.get_loads()

    else:
        return

    n = len(objects)
    p = len(properties)
    vals = np.zeros((n, p))
    extended_prop = np.zeros(p, dtype=object)
    log_scale_extended = np.zeros(p, dtype=object)
    for j in range(len(properties)):

        if len(objects):
            gc_prop = objects[0].registered_properties[properties[j]]

            for i, elem in enumerate(objects):
                val = elem.get_property_value(prop=gc_prop, t_idx=t_idx)
                vals[i, j] = val
                extended_prop[j] = properties[j]
                log_scale_extended[j] = log_scale[j]

    # create figure if needed
    if fig is None:
        fig = plt.figure(figsize=(12, 6))

    fig.suptitle('Analysis of the ' + str(object_type), fontsize=16)
    fig.set_facecolor('white')

    if n > 0:
        k = int(np.round(math.sqrt(p)))
        axs = np.empty(p + 1, dtype=object)

        for j in range(p):
            x = vals[:, j]
            mu = x.mean()
            variance = x.var()
            sigma = math.sqrt(variance)
            r = (mu - 6 * sigma, mu + 6 * sigma)

            # plot
            ax = fig.add_subplot(k, k + 1, j + 1)
            ax.set_facecolor('white')
            # bin_edges = np.histogram_bin_edges(x)
            ax.hist(x,
                    # bins=len(bin_edges),
                    range=r,
                    cumulative=False,
                    bottom=None,
                    histtype='bar',
                    align='mid',
                    orientation='vertical')
            ax.plot(x, np.zeros(n), 'o')
            ax.set_title(str(extended_prop[j]))

            if log_scale_extended[j]:
                ax.set_xscale('log')

            axs[j] = ax

        if object_type in [DeviceType.LineDevice.value,
                           DeviceType.Transformer2WDevice.value]:
            r = vals[:, 0]
            x = vals[:, 1]

            # plot
            ax = fig.add_subplot(k, k + 1, p + 2)
            ax.set_facecolor('white')
            ax.scatter(r, x)
            ax.set_title("R-X")
            ax.set_xlabel("R")
            ax.set_ylabel("X")
            axs[p] = ax

    fig.tight_layout(rect=(0, 0.03, 1, 0.95))
